@@ -101,37 +101,40 @@ typedef enum {
     /* Is an STable. */
     MVM_CF_STABLE = 2,
 
+    /* Is a heap-promoted call frame. */
+    MVM_CF_FRAME = 4,
+
     /* Has already been seen once in GC nursery. */
-    MVM_CF_NURSERY_SEEN = 4,
+    MVM_CF_NURSERY_SEEN = 8,
 
     /* Has been promoted to the old generation. */
-    MVM_CF_SECOND_GEN = 8,
+    MVM_CF_SECOND_GEN = 16,
 
     /* Is shared - that is, more than one thread knows about it. */
-    MVM_CF_SHARED = 16,
+    MVM_CF_SHARED = 32,
 
     /* Has already been added to the gen2 aggregates pointing to nursery
      * objects list. */
-    MVM_CF_IN_GEN2_ROOT_LIST = 32,
+    MVM_CF_IN_GEN2_ROOT_LIST = 64,
 
     /* A full GC run has found this object to be live. */
-    MVM_CF_GEN2_LIVE = 64,
+    MVM_CF_GEN2_LIVE = 128,
 
     /* This object in fromspace is live with a valid forwarder. */
     /* TODO - should be possible to use the same bit for this and GEN2_LIVE. */
-    MVM_CF_FORWARDER_VALID = 128,
+    MVM_CF_FORWARDER_VALID = 256,
 
     /* Have we allocated memory to store a serialization index? */
-    MVM_CF_SERIALZATION_INDEX_ALLOCATED = 256,
+    MVM_CF_SERIALZATION_INDEX_ALLOCATED = 512,
 
     /* Have we arranged a persistent object ID for this object? */
-    MVM_CF_HAS_OBJECT_ID = 512,
+    MVM_CF_HAS_OBJECT_ID = 1024,
 
     /* Have we flagged this object as something we must never repossess? */
     /* Note: if you're hunting for a flag, some day in the future when we
      * have used them all, this one is easy enough to eliminate by having the
      * tiny number of objects marked this way in a remembered set. */
-    MVM_CF_NEVER_REPOSSESS = 1024
+    MVM_CF_NEVER_REPOSSESS = 2048
 } MVMCollectableFlags;
 
 #ifdef MVM_USE_OVERFLOW_SERIALIZATION_INDEX
@@ -231,7 +234,7 @@ struct MVMBoolificationSpec {
 /* Constant for incrementing the type cache ID for new STables. This leaves
  * the lowest bits free for caches to attach flags (of note, the multi
  * dispatch cache). */
-#define MVM_TYPE_CACHE_ID_INCR 128
+#define MVM_TYPE_CACHE_ID_INCR 256
 
 /* S-table, representing a meta-object/representation pairing. Note that the
  * items are grouped in hope that it will pack decently and do decently in
@@ -342,6 +345,10 @@ struct MVMSTable {
     /* Also info we need to lazily deserialize the method cache. */
     MVMuint32                method_cache_offset;
     MVMSerializationContext *method_cache_sc;
+
+    /* A string associated with this STable for debugging purposes.
+     * Usually the name of the class this belongs to. */
+    char *debug_name;
 };
 
 /* The representation operations table. Note that representations are not
@@ -403,6 +410,16 @@ struct MVMREPROps_Boxing {
     MVMString * (*get_str) (MVMThreadContext *tc, MVMSTable *st,
         MVMObject *root, void *data);
 
+    /* Used with boxing. Sets an unsinged integer value, for representations
+     *  that can hold one. */
+    void (*set_uint) (MVMThreadContext *tc, MVMSTable *st,
+        MVMObject *root, void *data, MVMuint64 value);
+
+    /* Used with boxing. Gets an unsigned integer value, for representations
+     * that can hold one. */
+    MVMuint64 (*get_uint) (MVMThreadContext *tc, MVMSTable *st,
+        MVMObject *root, void *data);
+
     /* Some objects serve primarily as boxes of others, inlining them. This gets
      * gets the reference to such things, using the representation ID to distinguish
      * them. */
@@ -604,9 +621,13 @@ struct MVMREPROps {
     /* The representation's ID. */
     MVMuint32 ID;
 
-    /* Does this representation reference frames (either MVMStaticFrame or
-     * MVMFrame)? */
-    MVMuint32 refs_frames;
+    /* Optional API, for representations that allocate additonal memory and
+     * want to report its size for debugging purposes. */
+    MVMuint64 (*unmanaged_size) (MVMThreadContext *tc, MVMSTable *st, void *data);
+
+    /* Optional API to describe references to other Collectables either by
+     * index or by name, i.E. names of attributes or lexicals. */
+    void (*describe_refs) (MVMThreadContext *tc, MVMHeapSnapshotState *ss, MVMSTable *st, void *data);
 };
 
 /* Various handy macros for getting at important stuff. */
