@@ -42,6 +42,7 @@
 #include "6model/reprs/CPPStruct.h"
 #include "6model/reprs/NativeRef.h"
 #include "6model/reprs/MultiDimArray.h"
+#include "6model/reprs/Decoder.h"
 
 /* REPR related functions. */
 void MVM_repr_initialize_registry(MVMThreadContext *tc);
@@ -94,8 +95,9 @@ const MVMREPROps * MVM_repr_get_by_name(MVMThreadContext *tc, MVMString *name);
 #define MVM_REPR_ID_MVMCUnion               40
 #define MVM_REPR_ID_MultiDimArray           41
 #define MVM_REPR_ID_MVMCPPStruct            42
+#define MVM_REPR_ID_Decoder                 43
 
-#define MVM_REPR_CORE_COUNT                 43
+#define MVM_REPR_CORE_COUNT                 44
 #define MVM_REPR_MAX_COUNT                  64
 
 /* Default attribute functions for a REPR that lacks them. */
@@ -116,6 +118,8 @@ const MVMREPROps * MVM_repr_get_by_name(MVMThreadContext *tc, MVMString *name);
     MVM_REPR_DEFAULT_GET_NUM, \
     MVM_REPR_DEFAULT_SET_STR, \
     MVM_REPR_DEFAULT_GET_STR, \
+    MVM_REPR_DEFAULT_SET_UINT, \
+    MVM_REPR_DEFAULT_GET_UINT, \
     MVM_REPR_DEFAULT_GET_BOXED_REF \
 }
 
@@ -172,6 +176,8 @@ void MVM_REPR_DEFAULT_SET_NUM(MVMThreadContext *tc, MVMSTable *st, MVMObject *ro
 MVMnum64 MVM_REPR_DEFAULT_GET_NUM(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data);
 void MVM_REPR_DEFAULT_SET_STR(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMString *value);
 MVMString * MVM_REPR_DEFAULT_GET_STR(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data);
+void MVM_REPR_DEFAULT_SET_UINT(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMuint64 value);
+MVMuint64 MVM_REPR_DEFAULT_GET_UINT(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data);
 void * MVM_REPR_DEFAULT_GET_BOXED_REF(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMuint32 repr_id);
 
 /* Default positional indexing REPR function for a REPR that lacks it. */
