@@ -1,5 +1,4 @@
 #include "moar.h"
-#include <math.h>
 
 /* This representation's function pointer table. */
 static const MVMREPROps this_repr;
@@ -104,10 +103,14 @@ static MVMObject * index_mapping_and_flat_list(MVMThreadContext *tc, MVMObject *
     return flat_list;
 }
 
+static MVMint32 round_up_to_multi(MVMint32 i, MVMint32 m) {
+    return (MVMint32)((i + m - 1) / m) * m;
+}
+
 /* This works out an allocation strategy for the object. It takes care of
  * "inlining" storage of attributes that are natively typed, as well as
  * noting unbox targets. */
-static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_info, MVMCStructREPRData *repr_data) {
+static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_info, MVMCStructREPRData *repr_data, MVMSTable *st) {
     /* Compute index mapping table and get flat list of attributes. */
     MVMObject *flat_list = index_mapping_and_flat_list(tc, repr_info, repr_data);
 
@@ -121,17 +124,18 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
     else {
         /* We track the size of the struct, which is what we'll want offsets into. */
         MVMint32 cur_size    = 0;
-        /* The structure itself will be the multiple of its biggest element in size.
-         * So we keep track of that biggest element. */
-        MVMint32 multiple_of = 1;
+        MVMint32 struct_size = 0;
 
         /* Get number of attributes and set up various counters. */
         MVMint32 num_attrs        = MVM_repr_elems(tc, flat_list);
-        MVMint32 info_alloc       = num_attrs == 0 ? 1 : num_attrs;
+        MVMint32 info_alloc       = num_attrs;
         MVMint32 cur_obj_attr     = 0;
         MVMint32 cur_init_slot    = 0;
         MVMint32 i;
 
+        if (info_alloc == 0)
+            MVM_exception_throw_adhoc(tc, "Class %s has no attributes, which is illegal with the CStruct representation.", st->debug_name);
+
         /* Allocate location/offset arrays and GC mark info arrays. */
         repr_data->num_attributes      = num_attrs;
         repr_data->attribute_locations = (MVMint32 *)   MVM_malloc(info_alloc * sizeof(MVMint32));
@@ -236,7 +240,9 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                 }
                 else {
                     MVM_exception_throw_adhoc(tc,
-                        "CStruct representation only handles int, num, CArray, CPointer, CStruct, CPPStruct and CUnion");
+                        "CStruct representation only handles attributes of type:\n"
+                        "  (u)int8, (u)int16, (u)int32, (u)int64, (u)long, (u)longlong, num32, num64, (s)size_t, bool, Str\n"
+                        "  and types with representation: CArray, CPointer, CStruct, CPPStruct and CUnion");
                 }
             }
             else {
@@ -263,15 +269,12 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
             repr_data->struct_offsets[i] = cur_size;
             cur_size += bits / 8;
 
-            if (align > multiple_of)
-                multiple_of = align;
+            struct_size = round_up_to_multi(struct_size, align) + bits/8;
         }
 
         /* Finally, put computed allocation size in place; it's body size plus
          * header size. Also number of markables and sentinels. */
-        if (multiple_of > sizeof(void *))
-            multiple_of = sizeof(void *);
-        repr_data->struct_size = ceil((double)cur_size / (double)multiple_of) * multiple_of;
+        repr_data->struct_size = round_up_to_multi(struct_size, repr_data->struct_align);
         if (repr_data->initialize_slots)
             repr_data->initialize_slots[cur_init_slot] = -1;
     }
@@ -296,7 +299,7 @@ static MVMint32 try_get_slot(MVMThreadContext *tc, MVMCStructREPRData *repr_data
         while (cur_map_entry->class_key != NULL) {
             if (cur_map_entry->class_key == class_key) {
                 MVMObject *slot_obj = MVM_repr_at_key_o(tc, cur_map_entry->name_map, name);
-                if (IS_CONCRETE(slot_obj))
+                if (slot_obj && IS_CONCRETE(slot_obj))
                     return MVM_repr_get_int(tc, slot_obj);
                 break;
             }
@@ -325,7 +328,7 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *repr_info) {
     /* Compute allocation strategy. */
     MVMCStructREPRData *repr_data = MVM_calloc(1, sizeof(MVMCStructREPRData));
     MVMObject *attr_info = MVM_repr_at_key_o(tc, repr_info, tc->instance->str_consts.attribute);
-    compute_allocation_strategy(tc, attr_info, repr_data);
+    compute_allocation_strategy(tc, attr_info, repr_data, st);
     st->REPR_data = repr_data;
 }
 
