@@ -107,7 +107,13 @@ static void at_pos(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *d
 
     if (MVM_load(&cbq->elems) > 0) {
         MVMConcBlockingQueueNode *peeked;
-        uv_mutex_lock(&cbq->locks->head_lock);
+        MVMROOT(tc, root, {
+            MVM_gc_mark_thread_blocked(tc);
+            uv_mutex_lock(&cbq->locks->head_lock);
+            MVM_gc_mark_thread_unblocked(tc);
+            data = OBJECT_BODY(root);
+            cbq = (MVMConcBlockingQueueBody *)data;
+        });
         peeked = cbq->head->next;
         value->o = peeked ? peeked->value : tc->instance->VMNull;
         uv_mutex_unlock(&cbq->locks->head_lock);
@@ -126,6 +132,7 @@ static void push(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *dat
     MVMConcBlockingQueueBody *cbq = (MVMConcBlockingQueueBody *)data;
     MVMConcBlockingQueueNode *add;
     AO_t orig_elems;
+    MVMObject *to_add = value.o;
 
     if (kind != MVM_reg_obj)
         MVM_exception_throw_adhoc(tc,
@@ -135,16 +142,30 @@ static void push(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *dat
             "Cannot store a null value in a concurrent blocking queue");
 
     add = MVM_calloc(1, sizeof(MVMConcBlockingQueueNode));
-    MVM_ASSIGN_REF(tc, &(root->header), add->value, value.o);
 
-    uv_mutex_lock(&cbq->locks->tail_lock);
+    MVMROOT(tc, root, {
+    MVMROOT(tc, to_add, {
+        MVM_gc_mark_thread_blocked(tc);
+        uv_mutex_lock(&cbq->locks->tail_lock);
+        MVM_gc_mark_thread_unblocked(tc);
+    });
+    });
+    data = OBJECT_BODY(root);
+    cbq = (MVMConcBlockingQueueBody *)data;
+    MVM_ASSIGN_REF(tc, &(root->header), add->value, to_add);
     cbq->tail->next = add;
     cbq->tail = add;
     orig_elems = MVM_incr(&cbq->elems);
     uv_mutex_unlock(&cbq->locks->tail_lock);
 
     if (orig_elems == 0) {
-        uv_mutex_lock(&cbq->locks->head_lock);
+        MVMROOT(tc, root, {
+            MVM_gc_mark_thread_blocked(tc);
+            uv_mutex_lock(&cbq->locks->head_lock);
+            MVM_gc_mark_thread_unblocked(tc);
+        });
+        data = OBJECT_BODY(root);
+        cbq = (MVMConcBlockingQueueBody *)data;
         uv_cond_signal(&cbq->locks->head_cond);
         uv_mutex_unlock(&cbq->locks->head_lock);
     }
@@ -157,17 +178,21 @@ static void shift(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *da
     if (kind != MVM_reg_obj)
         MVM_exception_throw_adhoc(tc, "Can only shift objects from a ConcBlockingQueue");
 
-    uv_mutex_lock(&cbq->locks->head_lock);
+    MVMROOT(tc, root, {
+        MVM_gc_mark_thread_blocked(tc);
+        uv_mutex_lock(&cbq->locks->head_lock);
+        MVM_gc_mark_thread_unblocked(tc);
+        data = OBJECT_BODY(root);
+        cbq  = (MVMConcBlockingQueueBody *)data;
 
-    while (MVM_load(&cbq->elems) == 0) {
-        MVMROOT(tc, root, {
-            MVM_gc_mark_thread_blocked(tc);
-            uv_cond_wait(&cbq->locks->head_cond, &cbq->locks->head_lock);
-            MVM_gc_mark_thread_unblocked(tc);
-            data = OBJECT_BODY(root);
-            cbq  = (MVMConcBlockingQueueBody *)data;
-        });
-    }
+        while (MVM_load(&cbq->elems) == 0) {
+                MVM_gc_mark_thread_blocked(tc);
+                uv_cond_wait(&cbq->locks->head_cond, &cbq->locks->head_lock);
+                MVM_gc_mark_thread_unblocked(tc);
+                data = OBJECT_BODY(root);
+                cbq  = (MVMConcBlockingQueueBody *)data;
+        }
+    });
 
     taken = cbq->head->next;
     MVM_free(cbq->head);
@@ -243,7 +268,11 @@ MVMObject * MVM_concblockingqueue_poll(MVMThreadContext *tc, MVMConcBlockingQueu
     MVMConcBlockingQueueNode *taken;
     MVMObject *result = tc->instance->VMNull;
 
-    uv_mutex_lock(&cbq->body.locks->head_lock);
+    MVMROOT(tc, cbq, {
+        MVM_gc_mark_thread_blocked(tc);
+        uv_mutex_lock(&cbq->body.locks->head_lock);
+        MVM_gc_mark_thread_unblocked(tc);
+    });
 
     if (MVM_load(&cbq->body.elems) > 0) {
         taken = cbq->body.head->next;
