@@ -61,6 +61,9 @@ static void gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorkli
                 MVM_gc_worklist_add(tc, worklist, &ctx->args[i].o);
         }
     }
+    else {
+        MVM_gc_worklist_add(tc, worklist, &(body->use_mode_frame));
+    }
 }
 
 /* Called by the VM in order to free memory associated with this object. */
@@ -88,10 +91,6 @@ static void gc_free(MVMThreadContext *tc, MVMObject *obj) {
             MVM_free(ctx->body.apc);
         }
     }
-    else {
-        if (ctx->body.use_mode_frame)
-            MVM_frame_dec_ref(tc, ctx->body.use_mode_frame);
-    }
 }
 
 
@@ -145,5 +144,35 @@ static const MVMREPROps this_repr = {
     NULL, /* spesh */
     "MVMCallCapture", /* name */
     MVM_REPR_ID_MVMCallCapture,
-    1, /* refs_frames */
+    NULL, /* unmanaged_size */
+    NULL, /* describe_refs */
 };
+
+/* This function was only introduced for the benefit of the JIT. */
+MVMint64 MVM_capture_pos_primspec(MVMThreadContext *tc, MVMObject *obj, MVMint64 i) {
+    if (IS_CONCRETE(obj) && REPR(obj)->ID == MVM_REPR_ID_MVMCallCapture) {
+        MVMCallCapture *cc = (MVMCallCapture *)obj;
+        if (i >= 0 && i < cc->body.apc->num_pos) {
+            MVMCallsiteEntry *arg_flags = cc->body.apc->arg_flags
+                ? cc->body.apc->arg_flags
+                : cc->body.apc->callsite->arg_flags;
+            switch (arg_flags[i] & MVM_CALLSITE_ARG_MASK) {
+                case MVM_CALLSITE_ARG_INT:
+                    return MVM_STORAGE_SPEC_BP_INT;
+                case MVM_CALLSITE_ARG_NUM:
+                    return MVM_STORAGE_SPEC_BP_NUM;
+                case MVM_CALLSITE_ARG_STR:
+                    return MVM_STORAGE_SPEC_BP_STR;
+                default:
+                    return MVM_STORAGE_SPEC_BP_NONE;
+            }
+        }
+        else {
+            MVM_exception_throw_adhoc(tc,
+                "Bad argument index given to captureposprimspec");
+        }
+    }
+    else {
+        MVM_exception_throw_adhoc(tc, "captureposprimspec needs a MVMCallCapture");
+    }
+}
