@@ -345,7 +345,6 @@ static void deserialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
     MVMMultiDimArrayREPRData *repr_data = (MVMMultiDimArrayREPRData *)st->REPR_data;
     MVMMultiDimArrayBody     *body      = (MVMMultiDimArrayBody *)data;
     MVMint64 i, flat_elems;
-    size_t size;
 
     /* Read in dimensions. */
     for (i = 0; i < repr_data->num_dimensions; i++)
@@ -683,7 +682,7 @@ static void set_dimensions(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
     }
     else {
         MVM_exception_throw_adhoc(tc,
-            "Array type of %"PRId64" dimensions cannot be intialized with %"PRId64" dimensions",
+            "Array type of %"PRId64" dimensions cannot be initialized with %"PRId64" dimensions",
             repr_data->num_dimensions, num_dimensions);
     }
 }
