@@ -229,7 +229,7 @@ static void get_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
         MVM_exception_throw_adhoc(tc, "P6opaque: must compose before using get_attribute");
 
     /* Try the slot allocation first. */
-    slot = hint >= 0 && !(repr_data->mi) ? hint :
+    slot = hint >= 0 && hint < repr_data->num_attributes && !(repr_data->mi) ? hint :
         try_get_slot(tc, repr_data, class_handle, name);
     if (slot >= 0) {
         MVMSTable *attr_st = repr_data->flattened_stables[slot];
@@ -341,7 +341,7 @@ static void bind_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
         MVM_exception_throw_adhoc(tc, "P6opaque: must compose before using bind_attribute_boxed");
 
     /* Try the slot allocation first. */
-    slot = hint >= 0 && !(repr_data->mi) ? hint :
+    slot = hint >= 0 && hint < repr_data->num_attributes && !(repr_data->mi) ? hint :
         try_get_slot(tc, repr_data, class_handle, name);
     if (slot >= 0) {
         MVMSTable *attr_st = repr_data->flattened_stables[slot];
@@ -349,12 +349,14 @@ static void bind_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
         case MVM_reg_obj: {
             MVMObject *value = value_reg.o;
             if (attr_st) {
-                if (attr_st == STABLE(value))
+                MVMSTable *value_st = STABLE(value);
+                if (attr_st == value_st)
                     st->REPR->copy_to(tc, attr_st, OBJECT_BODY(value), root,
                         (char *)data + repr_data->attribute_offsets[slot]);
                 else
                     MVM_exception_throw_adhoc(tc,
-                        "P6opaque: representation mismatch when storing value to attribute");
+                        "P6opaque: representation mismatch when storing value (of type %s) to attribute (of type %s)",
+                        value_st->debug_name, attr_st->debug_name);
             }
             else {
                 set_obj_at_offset(tc, root, data, repr_data->attribute_offsets[slot], value);
@@ -402,8 +404,17 @@ static void bind_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
 /* Checks if an attribute has been initialized. */
 static MVMint64 is_attribute_initialized(MVMThreadContext *tc, MVMSTable *st, void *data, MVMObject *class_handle, MVMString *name, MVMint64 hint) {
     MVMP6opaqueREPRData *repr_data = (MVMP6opaqueREPRData *)st->REPR_data;
-    MVMint64 slot = try_get_slot(tc, repr_data, class_handle, name);
+    MVMint64 slot;
+
+    if (!repr_data)
+        MVM_exception_throw_adhoc(tc, "P6opaque: must compose before using bind_attribute_boxed");
+
     data = MVM_p6opaque_real_data(tc, data);
+    /* This can stay commented out until we actually pass something other than NO_HINT
+    slot = hint >= 0 && hint < repr_data->num_attributes && !(repr_data->mi) ? hint :
+        try_get_slot(tc, repr_data, class_handle, name);
+    */
+    slot = try_get_slot(tc, repr_data, class_handle, name);
     if (slot >= 0)
         return NULL != get_obj_at_offset(data, repr_data->attribute_offsets[slot]);
     else
@@ -709,7 +720,7 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
 
             /* Ensure we have a name. */
             if (MVM_is_null(tc, name_obj))
-                MVM_exception_throw_adhoc(tc, "P6opaque: missing attribute name for attribute %d", i);
+                MVM_exception_throw_adhoc(tc, "P6opaque: missing attribute name for attribute %"PRId64, i);
 
             if (REPR(name_obj)->ID == MVM_REPR_ID_MVMString) {
                 MVM_ASSIGN_REF(tc, &(st->header), name_map->names[i], (MVMString *)name_obj);
@@ -757,19 +768,19 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
                             case MVM_STORAGE_SPEC_BP_INT:
                                 if (repr_data->unbox_int_slot >= 0)
                                     MVM_exception_throw_adhoc(tc,
-                                        "Duplicate box_target for native int: attributes %d and %d", repr_data->unbox_int_slot, i);
+                                        "Duplicate box_target for native int: attributes %d and %"PRId64, repr_data->unbox_int_slot, i);
                                 repr_data->unbox_int_slot = cur_slot;
                                 break;
                             case MVM_STORAGE_SPEC_BP_NUM:
                                 if (repr_data->unbox_num_slot >= 0)
                                     MVM_exception_throw_adhoc(tc,
-                                        "Duplicate box_target for native num: attributes %d and %d", repr_data->unbox_num_slot, i);
+                                        "Duplicate box_target for native num: attributes %d and %"PRId64, repr_data->unbox_num_slot, i);
                                 repr_data->unbox_num_slot = cur_slot;
                                 break;
                             case MVM_STORAGE_SPEC_BP_STR:
                                 if (repr_data->unbox_str_slot >= 0)
                                     MVM_exception_throw_adhoc(tc,
-                                        "Duplicate box_target for native str: attributes %d and %d", repr_data->unbox_str_slot, i);
+                                        "Duplicate box_target for native str: attributes %d and %"PRId64, repr_data->unbox_str_slot, i);
                                 repr_data->unbox_str_slot = cur_slot;
                                 break;
                             default:
@@ -808,7 +819,7 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
             if (MVM_repr_exists_key(tc, attr_info, str_pos_del)) {
                 if (repr_data->pos_del_slot != -1)
                     MVM_exception_throw_adhoc(tc,
-                        "Duplicate positional delegate attributes: %d and %d", repr_data->pos_del_slot, cur_slot);
+                        "Duplicate positional delegate attributes: %d and %"PRId64"", repr_data->pos_del_slot, cur_slot);
                 if (unboxed_type == MVM_STORAGE_SPEC_BP_NONE)
                     repr_data->pos_del_slot = cur_slot;
                 else
@@ -818,7 +829,7 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
             if (MVM_repr_exists_key(tc, attr_info, str_ass_del)) {
                 if (repr_data->ass_del_slot != -1)
                     MVM_exception_throw_adhoc(tc,
-                        "Duplicate associative delegate attributes: %d and %d", repr_data->pos_del_slot, cur_slot);
+                        "Duplicate associative delegate attributes: %d and %"PRId64, repr_data->pos_del_slot, cur_slot);
                 if (unboxed_type == MVM_STORAGE_SPEC_BP_NONE)
                     repr_data->ass_del_slot = cur_slot;
                 else
@@ -1060,6 +1071,10 @@ static void deserialize_repr_data(MVMThreadContext *tc, MVMSTable *st, MVMSerial
             if (cur_st->REPR->gc_cleanup)
                 repr_data->gc_cleanup_slots[cur_gc_cleanup_slot++] = i;
 
+            if (spec->align == 0) {
+                MVM_exception_throw_adhoc(tc, "Serialization error: Storage Spec of P6opaque must not have align set to 0.");
+            }
+
             if (cur_offset % spec->align) {
                 cur_offset += spec->align - cur_offset % spec->align;
             }
@@ -1097,13 +1112,15 @@ static void deserialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
 /* Serializes the object's body. */
 static void serialize(MVMThreadContext *tc, MVMSTable *st, void *data, MVMSerializationWriter *writer) {
     MVMP6opaqueREPRData *repr_data = (MVMP6opaqueREPRData *)st->REPR_data;
-    MVMuint16 num_attributes = repr_data->num_attributes;
+    MVMuint16 num_attributes;
     MVMuint16 i;
 
     if (!repr_data)
         MVM_exception_throw_adhoc(tc,
             "Representation must be composed before it can be serialized");
 
+    num_attributes = repr_data->num_attributes;
+
     data = MVM_p6opaque_real_data(tc, data);
 
     for (i = 0; i < num_attributes; i++) {
