@@ -396,7 +396,7 @@ void MVM_args_assert_nameds_used(MVMThreadContext *tc, MVMArgProcContext *ctx) {
                     ctx->args[ctx->num_pos + 2 * i].s);
                 char *waste[] = { c_param, NULL };
                 MVM_exception_throw_adhoc_free(tc, waste,
-                    "Unexpected named parameter '%s' passed",
+                    "Unexpected named argument '%s' passed",
                     c_param);
             }
     }
@@ -781,7 +781,7 @@ static void flatten_args(MVMThreadContext *tc, MVMArgProcContext *ctx) {
                 unsigned bucket_tmp;
 
                 HASH_ITER(hash_handle, body->hash_head, current, tmp, bucket_tmp) {
-                    MVMString *arg_name = (MVMString *)current->key;
+                    MVMString *arg_name = MVM_HASH_KEY(current);
                     if (!seen_name(tc, arg_name, new_args, new_num_pos, new_arg_pos)) {
                         if (new_arg_pos + 1 >= new_args_size) {
                             new_args = MVM_realloc(new_args, (new_args_size *= 2) * sizeof(MVMRegister));
