@@ -283,11 +283,11 @@ static void deserialize_sc_deps(MVMThreadContext *tc, MVMCompUnit *cu, ReaderSta
 
         /* See if we can resolve it. */
         uv_mutex_lock(&tc->instance->mutex_sc_weakhash);
-        MVM_string_flatten(tc, handle);
         MVM_HASH_GET(tc, tc->instance->sc_weakhash, handle, scb);
         if (scb && scb->sc) {
             cu_body->scs_to_resolve[i] = NULL;
             MVM_ASSIGN_REF(tc, &(cu->common.header), cu_body->scs[i], scb->sc);
+            scb->claimed = 1;
         }
         else {
             if (!scb) {
@@ -625,7 +625,6 @@ void MVM_bytecode_finish_frame(MVMThreadContext *tc, MVMCompUnit *cu,
             entry->value = j;
 
             sf->body.lexical_types[j] = read_int16(pos, 6 * j);
-            MVM_string_flatten(tc, name);
             MVM_HASH_BIND(tc, sf->body.lexical_names, name, entry)
         }
         pos += 6 * sf->body.num_lexicals;
@@ -649,6 +648,7 @@ void MVM_bytecode_finish_frame(MVMThreadContext *tc, MVMCompUnit *cu,
                 sf->body.handlers[j].label_reg = read_int16(pos, 0);
                 pos += 2;
             }
+            sf->body.handlers[j].inlined_and_not_lexical = 0;
         }
     }
 
@@ -665,6 +665,12 @@ void MVM_bytecode_finish_frame(MVMThreadContext *tc, MVMCompUnit *cu,
     for (j = 0; j < slvs; j++) {
         MVMuint16 lex_idx = read_int16(pos, 0);
         MVMuint16 flags   = read_int16(pos, 2);
+
+        if (lex_idx >= sf->body.num_lexicals) {
+            MVM_reentrantmutex_unlock(tc, (MVMReentrantMutex *)cu->body.update_mutex);
+            MVM_exception_throw_adhoc(tc, "Lexical index out of bounds: %d > %d", lex_idx, sf->body.num_lexicals);
+        }
+
         sf->body.static_env_flags[lex_idx] = flags;
         if (flags == 2 && !dump_only) {
             /* State variable; need to resolve wval immediately. Other kinds
@@ -787,7 +793,7 @@ static MVMCallsite ** deserialize_callsites(MVMThreadContext *tc, MVMCompUnit *c
 
         if (rs->version >= 3 && nameds_non_flattening) {
             ensure_can_read(tc, cu, rs, pos, nameds_non_flattening * 4);
-            callsites[i]->arg_names = MVM_malloc(nameds_non_flattening * sizeof(MVMString));
+            callsites[i]->arg_names = MVM_malloc(nameds_non_flattening * sizeof(MVMString*));
             for (j = 0; j < nameds_non_flattening; j++) {
                 callsites[i]->arg_names[j] = get_heap_string(tc, cu, rs, pos, 0);
                 pos += 4;
