@@ -5,8 +5,6 @@ int MVM_dll_load(MVMThreadContext *tc, MVMString *name, MVMString *path) {
     char *cpath;
     DLLib *lib;
 
-    MVM_string_flatten(tc, name);
-
     uv_mutex_lock(&tc->instance->mutex_dll_registry);
 
     MVM_HASH_GET(tc, tc->instance->dll_registry, name, entry);
@@ -39,8 +37,11 @@ int MVM_dll_load(MVMThreadContext *tc, MVMString *name, MVMString *path) {
         entry->name = name;
         entry->refcount = 0;
 
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->name);
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->name,
+            "DLL name");
         MVM_HASH_BIND(tc, tc->instance->dll_registry, name, entry);
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->hash_handle.key,
+            "DLL name hash key");
     }
 
     entry->lib = lib;
@@ -55,7 +56,6 @@ int MVM_dll_free(MVMThreadContext *tc, MVMString *name) {
 
     uv_mutex_lock(&tc->instance->mutex_dll_registry);
 
-    MVM_string_flatten(tc, name);
     MVM_HASH_GET(tc, tc->instance->dll_registry, name, entry);
 
     if (!entry) {
@@ -91,7 +91,6 @@ MVMObject * MVM_dll_find_symbol(MVMThreadContext *tc, MVMString *lib,
 
     uv_mutex_lock(&tc->instance->mutex_dll_registry);
 
-    MVM_string_flatten(tc, lib);
     MVM_HASH_GET(tc, tc->instance->dll_registry, lib, entry);
 
     if (!entry) {
