@@ -128,11 +128,6 @@ struct MVMFrame {
 
     /* Linked list of any continuation tags we have. */
     MVMContinuationTag *continuation_tags;
-    
-    /* Linked MVMContext object, so we can track the
-     * serialization context and such. */
-    /* note: used atomically */
-    MVMObject *context_object;
 
     /* Cache for dynlex lookup; if the name is non-null, the cache is valid
      * and the register can be accessed directly to find the contextual. */
@@ -144,10 +139,6 @@ struct MVMFrame {
     MVMuint16 allocd_work;
     MVMuint16 allocd_env;
 
-    /* Flags that the caller chain should be kept in place after return or
-     * unwind; used to make sure we can get a backtrace after an exception. */
-    MVMuint8 keep_caller;
-
     /* Flags that the frame has been captured in a continuation, and as
      * such we should keep everything in place for multiple invocations. */
     MVMuint8 in_continuation;
@@ -212,6 +203,7 @@ void MVM_frame_unwind_to(MVMThreadContext *tc, MVMFrame *frame, MVMuint8 *abs_ad
 MVM_PUBLIC void MVM_frame_destory(MVMThreadContext *tc, MVMFrame *frame);
 MVM_PUBLIC MVMObject * MVM_frame_get_code_object(MVMThreadContext *tc, MVMCode *code);
 MVM_PUBLIC void MVM_frame_capturelex(MVMThreadContext *tc, MVMObject *code);
+MVM_PUBLIC void MVM_frame_capture_inner(MVMThreadContext *tc, MVMObject *code);
 MVM_PUBLIC MVMObject * MVM_frame_takeclosure(MVMThreadContext *tc, MVMObject *code);
 MVM_PUBLIC MVMObject * MVM_frame_vivify_lexical(MVMThreadContext *tc, MVMFrame *f, MVMuint16 idx);
 MVM_PUBLIC MVMRegister * MVM_frame_find_lexical_by_name(MVMThreadContext *tc, MVMString *name, MVMuint16 type);
