@@ -42,6 +42,10 @@ MVMThreadContext * MVM_tc_create(MVMInstance *instance) {
     tc->next_frame_nr = 0;
     tc->current_frame_nr = 0;
 
+    /* Initialize last_payload, so we can be sure it's never NULL and don't
+     * need to check. */
+    tc->last_payload = instance->VMNull;
+
     return tc;
 }
 
@@ -54,9 +58,10 @@ void MVM_tc_destroy(MVMThreadContext *tc) {
     /* We run once again (non-blocking) to eventually close filehandles. */
     uv_run(tc->loop, UV_RUN_NOWAIT);
 
-    /* Free the nursery. */
+    /* Free the nursery and finalization queue. */
     MVM_free(tc->nursery_fromspace);
     MVM_free(tc->nursery_tospace);
+    MVM_free(tc->finalizing);
 
     /* Destroy the second generation allocator. */
     MVM_gc_gen2_destroy(tc->instance, tc->gen2);
