@@ -24,7 +24,6 @@ static void add_work(MVMThreadContext *tc, MVMThreadContext *stolen) {
  * the run, and are not counted. Returns the count of threads that should be
  * added to the finished countdown. */
 static MVMuint32 signal_one_thread(MVMThreadContext *tc, MVMThreadContext *to_signal) {
-
     /* Loop here since we may not succeed first time (e.g. the status of the
      * thread may change between the two ways we try to twiddle it). */
     while (1) {
@@ -254,7 +253,8 @@ void MVM_gc_mark_thread_blocked(MVMThreadContext *tc) {
         if (MVM_load(&tc->gc_status) == MVMGCStatus_INTERRUPT)
             MVM_gc_enter_from_interrupt(tc);
         else
-            MVM_panic(MVM_exitcode_gcorch, "Invalid GC status observed; aborting");
+            MVM_panic(MVM_exitcode_gcorch,
+                "Invalid GC status observed while blocking thread; aborting");
     }
 }
 
@@ -271,8 +271,22 @@ void MVM_gc_mark_thread_unblocked(MVMThreadContext *tc) {
     }
 }
 
+/* Checks if a thread has marked itself as blocked. Considers that the GC may
+ * have stolen its work and marked it as such also. So what this really
+ * answers is, "did this thread mark itself blocked, and since then not mark
+ * itself unblocked", which is useful if you need to conditionally unblock
+ * or re-block. If the status changes from blocked to stolen or stolen to
+ * blocked between checking this and calling unblock, it's safe anyway since
+ * these cases are handled in MVM_gc_mark_thread_unblocked. Note that this
+ * relies on a thread itself only ever calling block/unblock. */
+MVMint32 MVM_gc_is_thread_blocked(MVMThreadContext *tc) {
+    AO_t gc_status = MVM_load(&(tc->gc_status));
+    return gc_status == MVMGCStatus_UNABLE ||
+           gc_status == MVMGCStatus_STOLEN;
+}
+
 static MVMint32 is_full_collection(MVMThreadContext *tc) {
-    MVMuint64 percent_growth, threshold;
+    MVMuint64 percent_growth;
     size_t rss, promoted;
 
     /* If it's below the absolute minimum, quickly return. */
@@ -281,7 +295,7 @@ static MVMint32 is_full_collection(MVMThreadContext *tc) {
         return 0;
 
     /* Otherwise, consider percentage of resident set size. */
-    if (uv_resident_set_memory(&rss) < 0)
+    if (uv_resident_set_memory(&rss) < 0 || rss == 0)
         rss = 50 * 1024 * 1024;
     percent_growth = (100 * promoted) / rss;
     return percent_growth >= MVM_GC_GEN2_THRESHOLD_PERCENT;
@@ -291,6 +305,11 @@ static void run_gc(MVMThreadContext *tc, MVMuint8 what_to_do) {
     MVMuint8   gen;
     MVMuint32  i, n;
 
+#if MVM_GC_DEBUG
+    if (tc->in_spesh)
+        MVM_panic(1, "Must not GC when in the specializer/JIT\n");
+#endif
+
     /* Decide nursery or full collection. */
     gen = tc->instance->gc_full_collect ? MVMGCGenerations_Both : MVMGCGenerations_Nursery;
 
@@ -495,12 +514,6 @@ void MVM_gc_enter_from_interrupt(MVMThreadContext *tc) {
 void MVM_gc_global_destruction(MVMThreadContext *tc) {
     char *nursery_tmp;
 
-    /* Must wait until we're the only thread... */
-    while (tc->instance->num_user_threads) {
-        GC_SYNC_POINT(tc);
-        MVM_platform_thread_yield();
-    }
-
     /* Fake a nursery collection run by swapping the semi-
      * space nurseries. */
     nursery_tmp = tc->nursery_fromspace;
