@@ -88,6 +88,8 @@ void MVM_gc_root_add_instance_roots_to_worklist(MVMThreadContext *tc, MVMGCWorkl
      * keys of it anyway... */
     HASH_ITER(hash_handle, tc->instance->sc_weakhash, current, tmp, bucket_tmp) {
         /* mark the string handle pointer iff it hasn't yet been resolved */
+        add_collectable(tc, worklist, snapshot, current->hash_handle.key,
+            "SC weakhash hash key");
         if (!current->sc)
             add_collectable(tc, worklist, snapshot, current->handle,
                 "SC weakhash unresolved handle");
@@ -97,6 +99,8 @@ void MVM_gc_root_add_instance_roots_to_worklist(MVMThreadContext *tc, MVMGCWorkl
     }
 
     HASH_ITER(hash_handle, tc->instance->loaded_compunits, current_lcun, tmp_lcun, bucket_tmp) {
+        add_collectable(tc, worklist, snapshot, current_lcun->hash_handle.key,
+            "Loaded compilation unit hash key");
         add_collectable(tc, worklist, snapshot, current_lcun->filename,
             "Loaded compilation unit filename");
     }
@@ -111,7 +115,7 @@ void MVM_gc_root_add_tc_roots_to_worklist(MVMThreadContext *tc, MVMGCWorklist *w
     MVMNativeCallbackCacheHead *current_cbceh, *tmp_cbceh;
     unsigned bucket_tmp;
 
-    /* Any active exception handlers. */
+    /* Any active exception handlers and payload. */
     MVMActiveHandler *cur_ah = tc->active_handlers;
     while (cur_ah != NULL) {
         add_collectable(tc, worklist, snapshot, cur_ah->ex_obj, "Active handler exception object");
@@ -119,6 +123,8 @@ void MVM_gc_root_add_tc_roots_to_worklist(MVMThreadContext *tc, MVMGCWorklist *w
             add_collectable(tc, worklist, snapshot, cur_ah->frame, "Active handler frame");
         cur_ah = cur_ah->next_handler;
     }
+    add_collectable(tc, worklist, snapshot, tc->last_payload,
+        "Last exception payload");
 
     /* The thread object. */
     add_collectable(tc, worklist, snapshot, tc->thread_obj, "Thread object");
@@ -155,6 +161,8 @@ void MVM_gc_root_add_tc_roots_to_worklist(MVMThreadContext *tc, MVMGCWorklist *w
     HASH_ITER(hash_handle, tc->native_callback_cache, current_cbceh, tmp_cbceh, bucket_tmp) {
         MVMint32 i;
         MVMNativeCallback *entry = current_cbceh->head;
+        add_collectable(tc, worklist, snapshot, current_cbceh->hash_handle.key,
+            "Native callback cache key");
         while (entry) {
             for (i = 0; i < entry->num_types; i++)
                 add_collectable(tc, worklist, snapshot, entry->types[i],
@@ -351,10 +359,6 @@ void MVM_gc_root_add_frame_roots_to_worklist(MVMThreadContext *tc, MVMGCWorklist
     MVM_gc_worklist_add(tc, worklist, &cur_frame->code_ref);
     MVM_gc_worklist_add(tc, worklist, &cur_frame->static_info);
 
-    /* Add any context object. */
-    if (cur_frame->context_object)
-        MVM_gc_worklist_add(tc, worklist, &cur_frame->context_object);
-
     /* Mark special return data, if needed. */
     if (cur_frame->special_return_data && cur_frame->mark_special_return_data)
         cur_frame->mark_special_return_data(tc, cur_frame, worklist);
