@@ -5,10 +5,8 @@
 
 #ifdef _WIN32
 #  define IS_SLASH(c)     ((c) == L'\\' || (c) == L'/')
-#  define IS_NOT_SLASH(c) ((c) != L'\\' && (c) != L'/')
 #else
 #  define IS_SLASH(c)     ((c) == '/')
-#  define IS_NOT_SLASH(c) ((c) != '/')
 #endif
 
 #ifdef _WIN32
@@ -32,62 +30,36 @@ static char * UnicodeToUTF8(const wchar_t *str)
      return result;
 }
 
-static int mkdir_p(wchar_t *pathname, MVMint64 mode) {
-    size_t len = wcslen(pathname);
+static int mkdir_p(MVMThreadContext *tc, wchar_t *pathname, MVMint64 mode) {
+    wchar_t *p = pathname, ch;
 #else
-static int mkdir_p(char *pathname, MVMint64 mode) {
-    size_t len = strlen(pathname);
-
+static int mkdir_p(MVMThreadContext *tc, char *pathname, MVMint64 mode) {
+    char *p = pathname, ch;
+    uv_fs_t req;
 #endif
-    ssize_t r;
-    char tmp;
+    int created = 0;
 
-    /* '/' always exists. */
-    if (len == 0)
-        return 0;
-
-    while (len > 0 && IS_SLASH(pathname[len - 1]))
-        len--;
-
-    tmp = pathname[len];
-    pathname[len] = '\0';
+    for (;; ++p)
+        if (!*p || IS_SLASH(*p)) {
+            ch = *p;
+            *p  = '\0';
 #ifdef _WIN32
-    r = CreateDirectoryW(pathname, NULL);
-
-    if (!r && GetLastError() == ERROR_PATH_NOT_FOUND)
+            if (CreateDirectoryW(pathname, NULL)) {
+                created = 1;
+            }
 #else
-    r = mkdir(pathname, mode);
-
-    if (r == -1 && errno == ENOENT)
+            if (uv_fs_stat(tc->loop, &req, pathname, NULL) <= 0) {
+                if (mkdir(pathname, mode) != -1) {
+                    created = 1;
+                }
+            }
 #endif
-    {
-        ssize_t _len = len - 1;
-        char _tmp;
-
-        while (_len >= 0 && IS_NOT_SLASH(pathname[_len]))
-            _len--;
-
-        _tmp = pathname[_len];
-        pathname[_len] = '\0';
-
-        r = mkdir_p(pathname, mode);
-
-        pathname[_len] = _tmp;
-
-#ifdef _WIN32
-        if(r) {
-            r = CreateDirectoryW(pathname, NULL);
+            if (!(*p = ch)) break;
         }
-#else
-        if(r == 0) {
-            r = mkdir(pathname, mode);
-        }
-#endif
-    }
 
-    pathname[len] = tmp;
+    if (!created) return -1;
 
-    return r;
+    return 0;
 }
 
 /* Create a directory recursively. */
@@ -120,7 +92,7 @@ void MVM_dir_mkdir(MVMThreadContext *tc, MVMString *path, MVMint64 mode) {
         wcscat(wpathname, abs_dirname);
     }
 
-    if (!mkdir_p(wpathname, mode)) {
+    if (mkdir_p(tc, wpathname, mode) == -1) {
         DWORD error = GetLastError();
         if (error != ERROR_ALREADY_EXISTS) {
             MVM_free(wpathname);
@@ -130,9 +102,10 @@ void MVM_dir_mkdir(MVMThreadContext *tc, MVMString *path, MVMint64 mode) {
     MVM_free(wpathname);
 #else
 
-    if (mkdir_p(pathname, mode) == -1 && errno != EEXIST) {
+    if (mkdir_p(tc, pathname, mode) == -1 && errno != EEXIST) {
+        int mkdir_error = errno;
         MVM_free(pathname);
-        MVM_exception_throw_adhoc(tc, "Failed to mkdir: %d", errno);
+        MVM_exception_throw_adhoc(tc, "Failed to mkdir: %d", mkdir_error);
     }
 
     MVM_free(pathname);
@@ -176,8 +149,9 @@ void MVM_dir_chdir(MVMThreadContext *tc, MVMString *dir) {
     char * const dirstring = MVM_string_utf8_c8_encode_C_string(tc, dir);
 
     if (uv_chdir((const char *)dirstring) != 0) {
+        int chdir_error = errno;
         MVM_free(dirstring);
-        MVM_exception_throw_adhoc(tc, "chdir failed: %s", uv_strerror(errno));
+        MVM_exception_throw_adhoc(tc, "chdir failed: %s", uv_strerror(chdir_error));
     }
 
     MVM_free(dirstring);
@@ -207,6 +181,7 @@ static void gc_free(MVMThreadContext *tc, MVMObject *h, void *d) {
 #ifdef _WIN32
         if (data->dir_name)
             MVM_free(data->dir_name);
+
         if (data->dir_handle)
             FindClose(data->dir_handle);
 #else
@@ -285,10 +260,11 @@ MVMObject * MVM_dir_open(MVMThreadContext *tc, MVMString *dirname) {
 #else
     char * const dir_name = MVM_string_utf8_c8_encode_C_string(tc, dirname);
     DIR * const dir_handle = opendir(dir_name);
+    int opendir_error = errno;
     MVM_free(dir_name);
 
     if (!dir_handle)
-        MVM_exception_throw_adhoc(tc, "Failed to open dir: %d", errno);
+        MVM_exception_throw_adhoc(tc, "Failed to open dir: %d", opendir_error);
 
     data->dir_handle = dir_handle;
 #endif
@@ -343,16 +319,17 @@ MVMString * MVM_dir_read(MVMThreadContext *tc, MVMObject *oshandle) {
         return tc->instance->str_consts.empty;
     }
 #else
-    struct dirent entry;
-    struct dirent *result;
-    int ret;
 
-    ret = readdir_r(data->dir_handle, &entry, &result);
+    struct dirent *entry;
+    errno = 0; /* must reset errno so we won't check old errno */
 
-    if (ret == 0) {
-        if (result == NULL)
-            return tc->instance->str_consts.empty;
-        return MVM_string_decode(tc, tc->instance->VMString, entry.d_name, strlen(entry.d_name), data->encoding);
+    entry = readdir(data->dir_handle);
+
+    if (errno == 0) {
+        MVMString *ret = (entry == NULL)
+                       ? tc->instance->str_consts.empty
+                       : MVM_string_decode(tc, tc->instance->VMString, entry->d_name, strlen(entry->d_name), data->encoding);
+        return ret;
     }
 
     MVM_exception_throw_adhoc(tc, "Failed to read dirhandle: %d", errno);
