@@ -171,13 +171,13 @@ MVMint64 MVM_proc_shell(MVMThreadContext *tc, MVMString *cmd, MVMString *cwd, MV
     uv_process_t *process = MVM_calloc(1, sizeof(uv_process_t));
     uv_process_options_t process_options = {0};
     uv_stdio_container_t process_stdio[3];
-    int i;
+    int i, process_still_running;
 
     char * const cmdin = MVM_string_utf8_c8_encode_C_string(tc, cmd);
     char * const _cwd = MVM_string_utf8_c8_encode_C_string(tc, cwd);
     const MVMuint64 size = MVM_repr_elems(tc, env);
-    MVMIter * const iter = (MVMIter *)MVM_iter(tc, env);
     char **_env = MVM_malloc((size + 1) * sizeof(char *));
+    MVMIter *iter;
 
 #ifdef _WIN32
     const MVMuint16 acp = GetACP(); /* We should get ACP at runtime. */
@@ -195,10 +195,19 @@ MVMint64 MVM_proc_shell(MVMThreadContext *tc, MVMString *cmd, MVMString *cwd, MV
     args[3] = NULL;
 #endif
 
-    INIT_ENV();
+    MVMROOT(tc, in, {
+    MVMROOT(tc, out, {
+    MVMROOT(tc, err, {
+        iter = (MVMIter *)MVM_iter(tc, env);
+        INIT_ENV();
+    });
+    });
+    });
+
     setup_process_stdio(tc, in,  process, &process_stdio[0], 0, flags,      "shell");
     setup_process_stdio(tc, out, process, &process_stdio[1], 1, flags >> 3, "shell");
-    setup_process_stdio(tc, err, process, &process_stdio[2], 2, flags >> 6, "shell");
+    if (!(flags & MVM_PIPE_MERGED_OUT_ERR))
+        setup_process_stdio(tc, err, process, &process_stdio[2], 2, flags >> 6, "shell");
 
     process_options.stdio       = process_stdio;
     process_options.file        = _cmd;
@@ -206,9 +215,14 @@ MVMint64 MVM_proc_shell(MVMThreadContext *tc, MVMString *cmd, MVMString *cwd, MV
     process_options.cwd         = _cwd;
     process_options.flags       = UV_PROCESS_WINDOWS_VERBATIM_ARGUMENTS | UV_PROCESS_WINDOWS_HIDE;
     process_options.env         = _env;
-    process_options.stdio_count = 3;
+    if (flags & MVM_PIPE_MERGED_OUT_ERR) {
+        process_options.stdio_count = 2;
+    }
+    else
+        process_options.stdio_count = 3;
     process_options.exit_cb     = spawn_on_exit;
     if (flags & (MVM_PIPE_CAPTURE_IN | MVM_PIPE_CAPTURE_OUT | MVM_PIPE_CAPTURE_ERR)) {
+        process_still_running = 1;
         process->data = MVM_calloc(1, sizeof(MVMint64));
         uv_ref((uv_handle_t *)process);
         spawn_result = uv_spawn(tc->loop, process, &process_options);
@@ -216,6 +230,7 @@ MVMint64 MVM_proc_shell(MVMThreadContext *tc, MVMString *cmd, MVMString *cwd, MV
             result = spawn_result;
     }
     else {
+        process_still_running = 0;
         process->data = &result;
         uv_ref((uv_handle_t *)process);
         spawn_result = uv_spawn(tc->loop, process, &process_options);
@@ -233,6 +248,9 @@ MVMint64 MVM_proc_shell(MVMThreadContext *tc, MVMString *cmd, MVMString *cwd, MV
     MVM_free(cmdin);
     uv_unref((uv_handle_t *)process);
 
+    if (!process_still_running)
+        MVM_free(process);
+
     return result;
 }
 
@@ -246,11 +264,11 @@ MVMint64 MVM_proc_spawn(MVMThreadContext *tc, MVMObject *argv, MVMString *cwd, M
 
     char   * const      _cwd = MVM_string_utf8_c8_encode_C_string(tc, cwd);
     const MVMuint64     size = MVM_repr_elems(tc, env);
-    MVMIter * const     iter = (MVMIter *)MVM_iter(tc, env);
     char              **_env = MVM_malloc((size + 1) * sizeof(char *));
     const MVMuint64  arg_size = MVM_repr_elems(tc, argv);
     char             **args = MVM_malloc((arg_size + 1) * sizeof(char *));
     MVMRegister        reg;
+    MVMIter           *iter;
 
     i = 0;
     while(i < arg_size) {
@@ -259,10 +277,19 @@ MVMint64 MVM_proc_spawn(MVMThreadContext *tc, MVMObject *argv, MVMString *cwd, M
     }
     args[arg_size] = NULL;
 
-    INIT_ENV();
+    MVMROOT(tc, in, {
+    MVMROOT(tc, out, {
+    MVMROOT(tc, err, {
+        iter = (MVMIter *)MVM_iter(tc, env);
+        INIT_ENV();
+    });
+    });
+    });
+
     setup_process_stdio(tc, in,  process, &process_stdio[0], 0, flags,      "spawn");
     setup_process_stdio(tc, out, process, &process_stdio[1], 1, flags >> 3, "spawn");
-    setup_process_stdio(tc, err, process, &process_stdio[2], 2, flags >> 6, "spawn");
+    if (!(flags & MVM_PIPE_MERGED_OUT_ERR))
+        setup_process_stdio(tc, err, process, &process_stdio[2], 2, flags >> 6, "spawn");
 
     process_options.stdio       = process_stdio;
     process_options.file        = arg_size ? args[0] : NULL;
@@ -270,7 +297,11 @@ MVMint64 MVM_proc_spawn(MVMThreadContext *tc, MVMObject *argv, MVMString *cwd, M
     process_options.cwd         = _cwd;
     process_options.flags       = UV_PROCESS_WINDOWS_HIDE;
     process_options.env         = _env;
-    process_options.stdio_count = 3;
+    if (flags & MVM_PIPE_MERGED_OUT_ERR) {
+        process_options.stdio_count = 2;
+    }
+    else
+        process_options.stdio_count = 3;
     process_options.exit_cb     = spawn_on_exit;
     if (flags & (MVM_PIPE_CAPTURE_IN | MVM_PIPE_CAPTURE_OUT | MVM_PIPE_CAPTURE_ERR)) {
         process->data = MVM_calloc(1, sizeof(MVMint64));
@@ -403,7 +434,7 @@ static void write_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
     /* Encode the string, or extract buf data. */
     if (wi->str_data) {
         MVMuint64 output_size_64;
-        output = MVM_string_utf8_encode(tc, wi->str_data, &output_size_64);
+        output = MVM_string_utf8_encode(tc, wi->str_data, &output_size_64, 1);
         output_size = (int)output_size_64;
     }
     else {
@@ -498,7 +529,9 @@ static MVMAsyncTask * write_str(MVMThreadContext *tc, MVMOSHandle *h, MVMObject
     task->body.data = wi;
 
     /* Hand the task off to the event loop. */
-    MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    MVMROOT(tc, task, {
+        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    });
 
     return task;
 }
@@ -540,7 +573,9 @@ static MVMAsyncTask * write_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObjec
     task->body.data = wi;
 
     /* Hand the task off to the event loop. */
-    MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    MVMROOT(tc, task, {
+        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    });
 
     return task;
 }
@@ -641,6 +676,7 @@ static const MVMIOOps proc_op_table = {
     NULL,
     NULL,
     NULL,
+    NULL,
     proc_async_gc_mark,
     NULL
 };
@@ -702,11 +738,14 @@ static void on_alloc(uv_handle_t *handle, size_t suggested_size, uv_buf_t *buf)
 static void async_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf, SpawnInfo *si,
                        MVMObject *callback, MVMDecodeStream *ds, MVMuint32 seq_number) {
     MVMThreadContext *tc  = si->tc;
-    MVMObject        *arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
-    MVMAsyncTask     *t   = (MVMAsyncTask *)MVM_repr_at_pos_o(tc,
-        tc->instance->event_loop_active, si->work_idx);
+    MVMObject *arr;
+    MVMAsyncTask *t;
+    MVMROOT(tc, callback, {
+        arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
+        t = (MVMAsyncTask *)MVM_repr_at_pos_o(tc, tc->instance->event_loop_active, si->work_idx);
+    });
     MVM_repr_push_o(tc, arr, callback);
-    if (nread > 0) {
+    if (nread >= 0) {
         MVMROOT(tc, t, {
         MVMROOT(tc, arr, {
             /* Push the sequence number. */
@@ -742,9 +781,9 @@ static void async_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf,
     else if (nread == UV_EOF) {
         MVMROOT(tc, t, {
         MVMROOT(tc, arr, {
-            MVMObject *minus_one = MVM_repr_box_int(tc,
-                tc->instance->boot_types.BOOTInt, -1);
-            MVM_repr_push_o(tc, arr, minus_one);
+            MVMObject *final = MVM_repr_box_int(tc,
+                tc->instance->boot_types.BOOTInt, seq_number);
+            MVM_repr_push_o(tc, arr, final);
             MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
             MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
         });
@@ -833,7 +872,7 @@ static void spawn_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
         pipe->data = si;
         process_stdio[1].flags       = UV_CREATE_PIPE | UV_WRITABLE_PIPE;
         process_stdio[1].data.stream = (uv_stream_t *)pipe;
-        si->ds_stdout                = MVM_string_decodestream_create(tc, MVM_encoding_type_utf8, 0);
+        si->ds_stdout                = MVM_string_decodestream_create(tc, MVM_encoding_type_utf8, 0, 1);
         stdout_pipe                  = pipe;
         stdout_cb                    = async_spawn_stdout_chars_read;
     }
@@ -856,7 +895,7 @@ static void spawn_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
         pipe->data = si;
         process_stdio[2].flags       = UV_CREATE_PIPE | UV_WRITABLE_PIPE;
         process_stdio[2].data.stream = (uv_stream_t *)pipe;
-        si->ds_stderr                = MVM_string_decodestream_create(tc, MVM_encoding_type_utf8, 0);
+        si->ds_stderr                = MVM_string_decodestream_create(tc, MVM_encoding_type_utf8, 0, 1);
         stderr_pipe                  = pipe;
         stderr_cb                    = async_spawn_stderr_chars_read;
     }
@@ -1073,7 +1112,9 @@ MVMObject * MVM_proc_spawn_async(MVMThreadContext *tc, MVMObject *queue, MVMObje
     });
 
     /* Hand the task off to the event loop. */
-    MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    MVMROOT(tc, handle, {
+        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+    });
 
     return (MVMObject *)handle;
 }
@@ -1087,7 +1128,7 @@ void MVM_proc_kill_async(MVMThreadContext *tc, MVMObject *handle_obj, MVMint64 s
             /* It's fine; send the kill by cancelling the task. */
             MVMIOAsyncProcessData *data = (MVMIOAsyncProcessData *)handle->body.data;
             data->signal = signal;
-            MVM_io_eventloop_cancel_work(tc, data->async_task);
+            MVM_io_eventloop_cancel_work(tc, data->async_task, NULL, NULL);
             return;
         }
     }
