@@ -56,7 +56,6 @@ static MVMint64 closefh(MVMThreadContext *tc, MVMOSHandle *h) {
 static void bind_stdio_handle(MVMThreadContext *tc, MVMOSHandle *h, uv_stdio_container_t *stdio,
         uv_process_t *process) {
     MVMIOSyncPipeData *data = (MVMIOSyncPipeData *)h->body.data;
-    data->process           = process;
     stdio->flags            = UV_INHERIT_STREAM;
     stdio->data.stream      = data->ss.handle;
 }
@@ -67,6 +66,15 @@ static void gc_free(MVMThreadContext *tc, MVMObject *h, void *d) {
     do_close(tc, data);
 }
 
+/* Get native file descriptor. */
+static MVMint64 mvm_fileno(MVMThreadContext *tc, MVMOSHandle *h) {
+    MVMIOSyncPipeData *data = (MVMIOSyncPipeData *)h->body.data;
+    uv_os_fd_t fd;
+    if (uv_fileno((uv_handle_t *)data->ss.handle, &fd) >= 0)
+        return (MVMint64)fd;
+    return -1;
+}
+
 /* IO ops table, populated with functions. */
 static const MVMIOClosable     closable      = { closefh };
 static const MVMIOEncodable    encodable     = { MVM_io_syncstream_set_encoding };
@@ -83,6 +91,7 @@ static const MVMIOSyncWritable sync_writable = { MVM_io_syncstream_write_str,
 static const MVMIOSeekable          seekable = { MVM_io_syncstream_seek,
                                                  MVM_io_syncstream_tell };
 static const MVMIOPipeable     pipeable      = { bind_stdio_handle };
+static const MVMIOIntrospection introspection = { NULL, mvm_fileno };
 static const MVMIOOps op_table = {
     &closable,
     &encodable,
@@ -90,11 +99,12 @@ static const MVMIOOps op_table = {
     &sync_writable,
     NULL,
     NULL,
+    NULL,
     &seekable,
     NULL,
     &pipeable,
     NULL,
-    NULL,
+    &introspection,
     NULL,
     gc_free
 };
@@ -107,6 +117,7 @@ MVMObject * MVM_io_syncpipe(MVMThreadContext *tc) {
     uv_pipe_init(tc->loop, handle, 0);
     data->ss.handle   = (uv_stream_t *)handle;
     data->ss.encoding = MVM_encoding_type_utf8;
+    data->ss.translate_newlines = 1;
     MVM_string_decode_stream_sep_default(tc, &(data->ss.sep_spec));
     result->body.ops  = &op_table;
     result->body.data = data;
