@@ -75,7 +75,9 @@ static MVMint32 read_to_buffer(MVMThreadContext *tc, MVMIOSyncStreamData *data,
         if (tc->loop != data->handle->loop) {
             MVM_exception_throw_adhoc(tc, "Tried to read() on a socket from outside its originating thread");
         }
+        MVM_gc_mark_thread_blocked(tc);
         uv_run(tc->loop, UV_RUN_DEFAULT);
+        MVM_gc_mark_thread_unblocked(tc);
         return 1;
     }
     else {
@@ -86,7 +88,8 @@ static MVMint32 read_to_buffer(MVMThreadContext *tc, MVMIOSyncStreamData *data,
 /* Ensures we have a decode stream, creating it if we're missing one. */
 static void ensure_decode_stream(MVMThreadContext *tc, MVMIOSyncStreamData *data) {
     if (!data->ds)
-        data->ds = MVM_string_decodestream_create(tc, data->encoding, 0);
+        data->ds = MVM_string_decodestream_create(tc, data->encoding, 0,
+            data->translate_newlines);
 }
 
 /* Reads a single line from the stream. May serve it from a buffer, if we
@@ -185,7 +188,8 @@ MVMint64 MVM_io_syncstream_write_str(MVMThreadContext *tc, MVMOSHandle *h, MVMSt
     uv_buf_t write_buf;
     int r;
 
-    output = MVM_string_encode(tc, str, 0, -1, &output_size, data->encoding, NULL);
+    output = MVM_string_encode(tc, str, 0, -1, &output_size, data->encoding, NULL,
+        data->translate_newlines ? MVM_TRANSLATE_NEWLINE_OUTPUT : 0);
     if (newline) {
         output = (char *)MVM_realloc(output, ++output_size);
         output[output_size - 1] = '\n';
@@ -260,6 +264,15 @@ static MVMint64 is_tty(MVMThreadContext *tc, MVMOSHandle *h) {
     return data->is_tty;
 }
 
+/* Get native file descriptor. */
+static MVMint64 mvm_fileno(MVMThreadContext *tc, MVMOSHandle *h) {
+    MVMIOSyncStreamData *data = (MVMIOSyncStreamData *)h->body.data;
+    uv_os_fd_t fd;
+    if (uv_fileno((uv_handle_t *)data->handle, &fd) >= 0)
+        return (MVMint64)fd;
+    return -1;
+}
+
 /* Operations aiding process spawning and I/O handling. */
 static void bind_stdio_handle(MVMThreadContext *tc, MVMOSHandle *h, uv_stdio_container_t *stdio,
         uv_process_t *process) {
@@ -272,14 +285,17 @@ static void bind_stdio_handle(MVMThreadContext *tc, MVMOSHandle *h, uv_stdio_con
 static void gc_free(MVMThreadContext *tc, MVMObject *h, void *d) {
     MVMIOSyncStreamData *data = (MVMIOSyncStreamData *)d;
     if (data) {
-        if (data->handle && not_std_handle(tc, h)) {
+        if (data->handle) {
             uv_close((uv_handle_t *)data->handle, NULL);
+            uv_run(tc->loop, UV_RUN_DEFAULT);
+            MVM_free(data->handle);
             data->handle = NULL;
         }
         if (data->ds) {
             MVM_string_decodestream_destory(tc, data->ds);
             data->ds = NULL;
         }
+        MVM_string_decode_stream_sep_destroy(tc, &(data->sep_spec));
         MVM_free(data);
     }
 }
@@ -301,7 +317,7 @@ static const MVMIOSeekable          seekable = { MVM_io_syncstream_seek,
                                                  MVM_io_syncstream_tell };
 static const MVMIOPipeable     pipeable      = { bind_stdio_handle };
 
-static const MVMIOPossiblyTTY possibly_tty         = { is_tty };
+static const MVMIOIntrospection introspection = { is_tty, mvm_fileno };
 
 static const MVMIOOps op_table = {
     &closable,
@@ -310,11 +326,12 @@ static const MVMIOOps op_table = {
     &sync_writable,
     NULL,
     NULL,
+    NULL,
     &seekable,
     NULL,
     &pipeable,
     NULL,
-    &possibly_tty,
+    &introspection,
     NULL,
     gc_free
 };
@@ -325,7 +342,8 @@ MVMObject * MVM_io_syncstream_from_uvstream(MVMThreadContext *tc, uv_stream_t *h
     MVMIOSyncStreamData * const data   = MVM_calloc(1, sizeof(MVMIOSyncStreamData));
     data->handle      = handle;
     data->encoding    = MVM_encoding_type_utf8;
-    data->is_tty = is_tty;
+    data->is_tty      = is_tty;
+    data->translate_newlines = 1;
     MVM_string_decode_stream_sep_default(tc, &(data->sep_spec));
     result->body.ops  = &op_table;
     result->body.data = data;
