@@ -32,6 +32,9 @@ typedef struct {
     /* callsite ID */
     unsigned short callsite_id;
 
+    /* the identifier for the callsite, to clean up later */
+    unsigned char *identifier;
+
     /* the uthash hash handle. */
     UT_hash_handle hash_handle;
 } CallsiteReuseEntry;
@@ -260,6 +263,8 @@ static void cleanup_frame(VM, FrameState *fs) {
 
 /* Cleans up all allocated memory related to this compilation. */
 static void cleanup_all(VM, WriterState *ws) {
+    CallsiteReuseEntry *current, *tmp;
+    unsigned bucket_tmp;
     if (ws->cur_frame)
         cleanup_frame(vm, ws->cur_frame);
     if (ws->scdep_seg)
@@ -274,6 +279,9 @@ static void cleanup_all(VM, WriterState *ws) {
         MVM_free(ws->bytecode_seg);
     if (ws->annotation_seg)
         MVM_free(ws->annotation_seg);
+    HASH_ITER(hash_handle, ws->callsite_reuse_head, current, tmp, bucket_tmp) {
+        MVM_free(current->identifier);
+    }
     MVM_HASH_DESTROY(hash_handle, CallsiteReuseEntry, ws->callsite_reuse_head);
     MVM_free(ws);
 }
@@ -321,18 +329,35 @@ static unsigned short type_to_local_type(VM, WriterState *ws, MASTNode *type) {
     if (ss->inlineable) {
         switch (ss->boxed_primitive) {
             case MVM_STORAGE_SPEC_BP_INT:
-                switch (ss->bits) {
-                    case 8:
-                        return MVM_reg_int8;
-                    case 16:
-                        return MVM_reg_int16;
-                    case 32:
-                        return MVM_reg_int32;
-                    case 64:
-                        return MVM_reg_int64;
-                    default:
-                        cleanup_all(vm, ws);
-                        DIE(vm, "Invalid int size for local/lexical");
+                if (ss->is_unsigned) {
+                    switch (ss->bits) {
+                        case 8:
+                            return MVM_reg_uint8;
+                        case 16:
+                            return MVM_reg_uint16;
+                        case 32:
+                            return MVM_reg_uint32;
+                        case 64:
+                            return MVM_reg_uint64;
+                        default:
+                            cleanup_all(vm, ws);
+                            DIE(vm, "Invalid int size for local/lexical");
+                    }
+                }
+                else {
+                    switch (ss->bits) {
+                        case 8:
+                            return MVM_reg_int8;
+                        case 16:
+                            return MVM_reg_int16;
+                        case 32:
+                            return MVM_reg_int32;
+                        case 64:
+                            return MVM_reg_int64;
+                        default:
+                            cleanup_all(vm, ws);
+                            DIE(vm, "Invalid int size for local/lexical");
+                    }
                 }
                 break;
             case MVM_STORAGE_SPEC_BP_NUM:
@@ -684,6 +709,7 @@ static unsigned short get_callsite_id(VM, WriterState *ws, MASTNode *flag_node,
     }
     entry = (CallsiteReuseEntry *)MVM_malloc(sizeof(CallsiteReuseEntry));
     entry->callsite_id = (unsigned short)ws->num_callsites;
+    entry->identifier = identifier;
     HASH_ADD_KEYPTR(hash_handle, ws->callsite_reuse_head, identifier, identifier_len, entry);
 
     /* Emit callsite; be sure to pad if there's uneven number of flags. */
@@ -1330,8 +1356,8 @@ static char * form_string_heap(VM, WriterState *ws, unsigned int *string_heap_si
 
         /* Encode it with the chosen algorithm. */
         encoded = need_utf8
-            ? MVM_string_utf8_encode(tc, str, &bytelen)
-            : MVM_string_latin1_encode(tc, str, &bytelen);
+            ? MVM_string_utf8_encode(tc, str, &bytelen, 0)
+            : MVM_string_latin1_encode(tc, str, &bytelen, 0);
         if (bytelen > 0x3FFFFFFF) {
             cleanup_all(vm, ws);
             DIE(vm, "String too long for string constants segment");
@@ -1369,8 +1395,8 @@ static char * form_string_heap(VM, WriterState *ws, unsigned int *string_heap_si
 
 /* Takes all the pieces and forms the bytecode output. */
 static char * form_bytecode_output(VM, WriterState *ws, unsigned int *bytecode_size) {
-    unsigned int size    = 0;
-    unsigned int pos     = 0;
+    MVMuint32     size    = 0;
+    MVMuint32     pos     = 0;
     char         *output;
     unsigned int  string_heap_size;
     char         *string_heap;
@@ -1386,16 +1412,16 @@ static char * form_bytecode_output(VM, WriterState *ws, unsigned int *bytecode_s
     string_heap = form_string_heap(vm, ws, &string_heap_size);
 
     /* Work out total size. */
-    size += HEADER_SIZE;
-    size += string_heap_size;
-    size += ws->scdep_bytes;
-    size += ws->extops_bytes;
-    size += ws->frame_pos;
-    size += ws->callsite_pos;
-    size += ws->bytecode_pos;
-    size += ws->annotation_pos;
+    size += MVM_ALIGN_SECTION(HEADER_SIZE);
+    size += MVM_ALIGN_SECTION(string_heap_size);
+    size += MVM_ALIGN_SECTION(ws->scdep_bytes);
+    size += MVM_ALIGN_SECTION(ws->extops_bytes);
+    size += MVM_ALIGN_SECTION(ws->frame_pos);
+    size += MVM_ALIGN_SECTION(ws->callsite_pos);
+    size += MVM_ALIGN_SECTION(ws->bytecode_pos);
+    size += MVM_ALIGN_SECTION(ws->annotation_pos);
     if (vm->serialized)
-        size += vm->serialized_size;
+        size += MVM_ALIGN_SECTION(vm->serialized_size);
 
     /* Allocate space for the bytecode output. */
     output = (char *)MVM_malloc(size);
@@ -1404,37 +1430,37 @@ static char * form_bytecode_output(VM, WriterState *ws, unsigned int *bytecode_s
     /* Generate start of header. */
     memcpy(output, "MOARVM\r\n", 8);
     write_int32(output, 8, BYTECODE_VERSION);
-    pos += HEADER_SIZE;
+    pos += MVM_ALIGN_SECTION(HEADER_SIZE);
 
     /* Add SC dependencies section and its header entries. */
     write_int32(output, SCDEP_HEADER_OFFSET, pos);
     write_int32(output, SCDEP_HEADER_OFFSET + 4, ELEMS(vm, ws->cu->sc_handles));
     memcpy(output + pos, ws->scdep_seg, ws->scdep_bytes);
-    pos += ws->scdep_bytes;
+    pos += MVM_ALIGN_SECTION(ws->scdep_bytes);
 
     /* Add extension ops section and its header entries. */
     write_int32(output, EXTOP_HEADER_OFFSET, pos);
     write_int32(output, EXTOP_HEADER_OFFSET + 4, ws->num_extops);
     memcpy(output + pos, ws->extops_seg, ws->extops_bytes);
-    pos += ws->extops_bytes;
+    pos += MVM_ALIGN_SECTION(ws->extops_bytes);
 
     /* Add frames section and its header entries. */
     write_int32(output, FRAME_HEADER_OFFSET, pos);
     write_int32(output, FRAME_HEADER_OFFSET + 4, ws->num_frames);
     memcpy(output + pos, ws->frame_seg, ws->frame_pos);
-    pos += ws->frame_pos;
+    pos += MVM_ALIGN_SECTION(ws->frame_pos);
 
     /* Add callsites section and its header entries. */
     write_int32(output, CALLSITE_HEADER_OFFSET, pos);
     write_int32(output, CALLSITE_HEADER_OFFSET + 4, ws->num_callsites);
     memcpy(output + pos, ws->callsite_seg, ws->callsite_pos);
-    pos += ws->callsite_pos;
+    pos += MVM_ALIGN_SECTION(ws->callsite_pos);
 
     /* Add strings heap section and its header entries. */
     write_int32(output, STRING_HEADER_OFFSET, pos);
     write_int32(output, STRING_HEADER_OFFSET + 4, ELEMS(vm, ws->strings));
     memcpy(output + pos, string_heap, string_heap_size);
-    pos += string_heap_size;
+    pos += MVM_ALIGN_SECTION(string_heap_size);
     if (string_heap) {
         MVM_free(string_heap);
         string_heap = NULL;
@@ -1445,7 +1471,7 @@ static char * form_bytecode_output(VM, WriterState *ws, unsigned int *bytecode_s
         write_int32(output, SCDATA_HEADER_OFFSET, pos);
         write_int32(output, SCDATA_HEADER_OFFSET + 4, vm->serialized_size);
         memcpy(output + pos, vm->serialized, vm->serialized_size);
-        pos += vm->serialized_size;
+        pos += MVM_ALIGN_SECTION(vm->serialized_size);
         MVM_free(vm->serialized);
         vm->serialized = NULL;
         vm->serialized_size = 0;
@@ -1455,13 +1481,13 @@ static char * form_bytecode_output(VM, WriterState *ws, unsigned int *bytecode_s
     write_int32(output, BYTECODE_HEADER_OFFSET, pos);
     write_int32(output, BYTECODE_HEADER_OFFSET + 4, ws->bytecode_pos);
     memcpy(output + pos, ws->bytecode_seg, ws->bytecode_pos);
-    pos += ws->bytecode_pos;
+    pos += MVM_ALIGN_SECTION(ws->bytecode_pos);
 
     /* Add annotation section and its header entries (offset, length). */
     write_int32(output, ANNOTATION_HEADER_OFFSET, pos);
     write_int32(output, ANNOTATION_HEADER_OFFSET + 4, ws->annotation_pos);
     memcpy(output + pos, ws->annotation_seg, ws->annotation_pos);
-    pos += ws->annotation_pos;
+    pos += MVM_ALIGN_SECTION(ws->annotation_pos);
 
     /* Add HLL and special frame indexes. */
     write_int32(output, HLL_NAME_HEADER_OFFSET, hll_str_idx);
