@@ -5,27 +5,27 @@
 #include <float.h>
 #endif
 
-#ifdef INFINITY
+#if defined(INFINITY) && !defined(_AIX)
 static const MVMnum64 MVM_NUM_POSINF =  INFINITY;
 static const MVMnum64 MVM_NUM_NEGINF = -INFINITY;
 #else
-#ifdef _MSC_VER
+#  ifdef _MSC_VER
 #define MVM_NUM_POSINF  (DBL_MAX+DBL_MAX)
 #define MVM_NUM_NEGINF -(DBL_MAX+DBL_MAX)
-#else
+#  else
 static const MVMnum64 MVM_NUM_POSINF =  1.0 / 0.0;
 static const MVMnum64 MVM_NUM_NEGINF = -1.0 / 0.0;
-#endif
+#  endif
 #endif
 
-#ifdef NAN
+#if defined(NAN) && !defined(_AIX)
 static const MVMnum64 MVM_NUM_NAN = NAN;
 #else
-#ifdef _MSC_VER
+#  ifdef _MSC_VER
 #define MVM_NUM_NAN (MVM_NUM_POSINF-MVM_NUM_POSINF)
-#else
+#  else
 static const MVMnum64 MVM_NUM_NAN = 0.0 / 0.0;
-#endif
+#  endif
 #endif
 
 MVMint64 MVM_num_isnanorinf(MVMThreadContext *tc, MVMnum64 n) {
