@@ -66,7 +66,8 @@ static FILE *fopen_perhaps_with_pid(char *path, const char *mode) {
 /* Create a new instance of the VM. */
 MVMInstance * MVM_vm_create_instance(void) {
     MVMInstance *instance;
-    char *spesh_log, *spesh_nodelay, *spesh_disable, *spesh_inline_disable, *spesh_osr_disable;
+    char *spesh_log, *spesh_nodelay, *spesh_disable, *spesh_inline_disable,
+         *spesh_osr_disable, *spesh_limit;
     char *jit_log, *jit_disable, *jit_bytecode_dir;
     char *dynvar_log;
     int init_stat;
@@ -78,9 +79,7 @@ MVMInstance * MVM_vm_create_instance(void) {
     instance->main_thread = MVM_tc_create(instance);
     instance->main_thread->thread_id = 1;
 
-    /* No user threads when we start, and next thread to be created gets ID 2
-     * (the main thread got ID 1). */
-    instance->num_user_threads    = 0;
+    /* Next thread to be created gets ID 2 (the main thread got ID 1). */
     MVM_store(&instance->next_user_thread_id, 2);
 
     /* Set up the permanent roots storage. */
@@ -132,8 +131,9 @@ MVMInstance * MVM_vm_create_instance(void) {
     /* Bootstrap 6model. It is assumed the GC will not be called during this. */
     MVM_6model_bootstrap(instance->main_thread);
 
-    /* Fix up main thread's usecapture. */
+    /* Fix up main thread's usecapture and last_payload. */
     instance->main_thread->cur_usecapture = MVM_repr_alloc_init(instance->main_thread, instance->CallCapture);
+    instance->main_thread->last_payload = instance->VMNull;
 
     /* Initialize event loop thread starting mutex. */
     init_mutex(instance->mutex_event_loop_start, "event loop thread start");
@@ -206,6 +206,12 @@ MVMInstance * MVM_vm_create_instance(void) {
         instance->spesh_nodelay = 1;
     }
 
+    /* Should we limit the number of specialized frames produced? (This is
+     * mostly useful for building spesh bug bisect tools.) */
+    spesh_limit = getenv("MVM_SPESH_LIMIT");
+    if (spesh_limit && strlen(spesh_limit))
+        instance->spesh_limit = atoi(spesh_limit);
+
     /* JIT environment/logging setup. */
     jit_disable = getenv("MVM_JIT_DISABLE");
     if (!jit_disable || strlen(jit_disable) == 0)
