@@ -4,8 +4,18 @@
  * lexicals. */
 static void calculate_work_env_sizes(MVMThreadContext *tc, MVMStaticFrame *sf,
                                      MVMSpeshCandidate *c) {
-    c->work_size = (c->num_locals + sf->body.cu->body.max_callsite_size)
-        * sizeof(MVMRegister);
+    MVMuint32 max_callsite_size;
+    MVMint32 i;
+
+    max_callsite_size = sf->body.cu->body.max_callsite_size;
+
+    for (i = 0; i < c->num_inlines; i++) {
+        MVMuint32 cs = c->inlines[i].code->body.sf->body.cu->body.max_callsite_size;
+        if (cs > max_callsite_size)
+            max_callsite_size = cs;
+    }
+
+    c->work_size = (c->num_locals + max_callsite_size) * sizeof(MVMRegister);
     c->env_size = c->num_lexicals * sizeof(MVMRegister);
 }
 
@@ -27,13 +37,22 @@ MVMSpeshCandidate * MVM_spesh_candidate_setup(MVMThreadContext *tc,
     char *after = 0;
     MVMSpeshGraph *sg;
 
-    /* If we're profiling, log we're starting spesh work. */
+    /* If we've reached our specialization limit, don't continue. */
+    if (tc->instance->spesh_limit)
+        if (++tc->instance->spesh_produced > tc->instance->spesh_limit)
+            return NULL;
+
+    /* If we're profiling, log we're starting spesh work. Also track entry
+     * to spesh when GC debugging. */
     if (tc->instance->profiling)
         MVM_profiler_log_spesh_start(tc);
+#if MVM_GC_DEBUG
+    tc->in_spesh = 1;
+#endif
 
     /* Do initial generation of the specialization, working out the argument
      * guards and adding logging. */
-    sg = MVM_spesh_graph_create(tc, static_frame, 0);
+    sg = MVM_spesh_graph_create(tc, static_frame, 0, 1);
     if (tc->instance->spesh_log_fh)
         before = MVM_spesh_dump(tc, sg);
     MVM_spesh_args(tc, sg, callsite, args);
@@ -127,9 +146,12 @@ MVMSpeshCandidate * MVM_spesh_candidate_setup(MVMThreadContext *tc,
     }
     uv_mutex_unlock(&tc->instance->mutex_spesh_install);
 
-    /* If we're profiling, log we've finished spesh work. */
+    /* If we're profiling or GC debugging, log we've finished spesh work. */
     if (tc->instance->profiling)
         MVM_profiler_log_spesh_end(tc);
+#if MVM_GC_DEBUG
+    tc->in_spesh = 0;
+#endif
 
     MVM_free(sc);
     return result;
@@ -145,9 +167,12 @@ void MVM_spesh_candidate_specialize(MVMThreadContext *tc, MVMStaticFrame *static
     MVMSpeshGraph *sg;
     MVMJitGraph   *jg = NULL;
 
-    /* If we're profiling, log we're starting spesh work. */
+    /* If we're profiling or GC debugging, log we're starting spesh work. */
     if (tc->instance->profiling)
         MVM_profiler_log_spesh_start(tc);
+#if MVM_GC_DEBUG
+    tc->in_spesh = 1;
+#endif
 
     /* Obtain the graph, add facts, and do optimization work. */
     sg = candidate->sg;
@@ -219,9 +244,12 @@ void MVM_spesh_candidate_specialize(MVMThreadContext *tc, MVMStaticFrame *static
     MVM_barrier();
     candidate->sg = NULL;
 
-    /* If we're profiling, log we've finished spesh work. */
+    /* If we're profiling or GC debugging, log we've finished spesh work. */
     if (tc->instance->profiling)
         MVM_profiler_log_spesh_end(tc);
+#if MVM_GC_DEBUG
+    tc->in_spesh = 0;
+#endif
 }
 
 
