@@ -18,8 +18,8 @@ static void copy_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 to_orig
 
 /* Called when one set of facts depend on another, allowing any log guard
  * that is to thank to be marked used as needed later on. */
-static void depend(MVMThreadContext *tc, MVMSpeshGraph *g,
-                   MVMSpeshFacts *target, MVMSpeshFacts *source) {
+void MVM_spesh_facts_depend(MVMThreadContext *tc, MVMSpeshGraph *g,
+                            MVMSpeshFacts *target, MVMSpeshFacts *source) {
     if (source->flags & MVM_SPESH_FACT_FROM_LOG_GUARD) {
         target->flags     |= MVM_SPESH_FACT_FROM_LOG_GUARD;
         target->log_guard  = source->log_guard;
@@ -36,7 +36,7 @@ static void create_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 obj_o
     if (type_facts->flags & MVM_SPESH_FACT_KNOWN_TYPE) {
         obj_facts->type   = type_facts->type;
         obj_facts->flags |= MVM_SPESH_FACT_KNOWN_TYPE;
-        depend(tc, g, obj_facts, type_facts);
+        MVM_spesh_facts_depend(tc, g, obj_facts, type_facts);
     }
 
     /* We know it's a concrete object. */
@@ -95,9 +95,48 @@ static void object_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 tgt_o
     }
 }
 
+/* Checks if there's a possible aliasing operation that could cause the
+ * facts about the contents of a container to be invalid by the instruction
+ * under consideration. Assumes the instruction is a decont with argument 1
+ * being the thing to decontainerize. */
+MVMint32 MVM_spesh_facts_decont_blocked_by_alias(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *ins) {
+    /* No facts or no writer means we can't know it's safe; block. */
+    MVMSpeshFacts *facts = MVM_spesh_get_facts(tc, g, ins->operands[1]);
+    if (!facts || !facts->writer)
+        return 1;
+
+    /* Walk backwards over instructions. */
+    while (ins->prev) {
+        ins = ins->prev;
+
+        /* If we found the writer without anything blocking, we're good. */
+        if (ins == facts->writer)
+            return 0;
+
+        /* If there's an operation that may alias, blocked. */
+        switch (ins->info->opcode) {
+            case MVM_OP_bindattr_o:
+            case MVM_OP_bindattrs_o:
+            case MVM_OP_assign:
+            case MVM_OP_assignunchecked:
+            case MVM_OP_assign_i:
+            case MVM_OP_assign_n:
+            case MVM_OP_assign_s:
+            case MVM_OP_sp_bind_o:
+            case MVM_OP_sp_p6obind_o:
+                return 1;
+        }
+    }
+
+    /* We didn't find the writer in this basic block, meaning an invocation
+     * may have made it unsafe. Block. */
+    return 1;
+}
+
 /* Propagates information relating to decontainerization. */
-static void decont_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 out_orig,
-                         MVMuint16 out_i, MVMuint16 in_orig, MVMuint16 in_i) {
+static void decont_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *ins,
+                         MVMuint16 out_orig, MVMuint16 out_i, MVMuint16 in_orig,
+                         MVMuint16 in_i) {
     MVMSpeshFacts *out_facts = &(g->facts[out_orig][out_i]);
     MVMSpeshFacts *in_facts  = &(g->facts[in_orig][in_i]);
 
@@ -112,18 +151,20 @@ static void decont_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 out_o
 
     /* We may also know the original was containerized, and have some facts
      * about its contents. */
-    if (in_flags & MVM_SPESH_FACT_KNOWN_DECONT_TYPE) {
-        out_facts->type = in_facts->decont_type;
-        out_facts->flags |= MVM_SPESH_FACT_KNOWN_TYPE;
+    if (!MVM_spesh_facts_decont_blocked_by_alias(tc, g, ins)) {
+        if (in_flags & MVM_SPESH_FACT_KNOWN_DECONT_TYPE) {
+            out_facts->type = in_facts->decont_type;
+            out_facts->flags |= MVM_SPESH_FACT_KNOWN_TYPE;
+        }
+        if (in_flags & MVM_SPESH_FACT_DECONT_CONCRETE)
+            out_facts->flags |= MVM_SPESH_FACT_CONCRETE;
+        else if (in_flags & MVM_SPESH_FACT_DECONT_TYPEOBJ)
+            out_facts->flags |= MVM_SPESH_FACT_TYPEOBJ;
+        if (in_flags & (MVM_SPESH_FACT_KNOWN_DECONT_TYPE |
+                        MVM_SPESH_FACT_DECONT_CONCRETE |
+                        MVM_SPESH_FACT_DECONT_TYPEOBJ))
+            MVM_spesh_facts_depend(tc, g, out_facts, in_facts);
     }
-    if (in_flags & MVM_SPESH_FACT_DECONT_CONCRETE)
-        out_facts->flags |= MVM_SPESH_FACT_CONCRETE;
-    else if (in_flags & MVM_SPESH_FACT_DECONT_TYPEOBJ)
-        out_facts->flags |= MVM_SPESH_FACT_TYPEOBJ;
-    if (in_flags & (MVM_SPESH_FACT_KNOWN_DECONT_TYPE |
-                    MVM_SPESH_FACT_DECONT_CONCRETE |
-                    MVM_SPESH_FACT_DECONT_TYPEOBJ))
-        depend(tc, g, out_facts, in_facts);
 }
 
 /* Looks up a wval and adds information based on it. */
@@ -133,7 +174,7 @@ static void wval_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 tgt_ori
     if (dep < cu->body.num_scs) {
         MVMSerializationContext *sc = MVM_sc_get_sc(tc, cu, dep);
         if (sc)
-            object_facts(tc, g, tgt_orig, tgt_i, MVM_sc_get_object(tc, sc, idx));
+            object_facts(tc, g, tgt_orig, tgt_i, MVM_sc_try_get_object(tc, sc, idx));
     }
 }
 
@@ -192,7 +233,8 @@ static void literal_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *i
             tgt_facts->value.i = ins->operands[1].lit_i16;
             break;
         case MVM_OP_const_s:
-            tgt_facts->value.s   = g->sf->body.cu->body.strings[ins->operands[1].lit_str_idx];
+            tgt_facts->value.s = MVM_cu_string(tc, g->sf->body.cu,
+                ins->operands[1].lit_str_idx);
             break;
         default:
             return;
@@ -275,39 +317,59 @@ static void log_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb, MV
     /* Produce a guard op and set facts. */
     if (stable_cont) {
         MVMSpeshOperand reg  = ins->operands[0];
+        MVMContainerSpec *cs = (MVMContainerSpec *) STABLE(stable_cont)->container_spec;
         facts                = &g->facts[reg.reg.orig][reg.reg.i];
         facts->type          = STABLE(stable_cont)->WHAT;
         facts->flags        |= (MVM_SPESH_FACT_KNOWN_TYPE | MVM_SPESH_FACT_CONCRETE |
                                MVM_SPESH_FACT_KNOWN_DECONT_TYPE);
         facts->decont_type   = STABLE(stable_value)->WHAT;
-        if (STABLE(stable_cont)->container_spec->can_store(tc, stable_cont)) {
-            /* We could do stability testing on rw-ness too, but it's quite
-             * unlikely we'll have codepaths with a mix of readable and
-             * writable containers. */
+
+        /* If this is a native container, we get away with testing
+         * against the STABLE only, as the NativeRef REPR has all
+         * interesting values in its REPRData. */
+        if (cs->can_store(tc, stable_cont) &&
+                (MVM_6model_container_iscont_i(tc, stable_cont) ||
+                MVM_6model_container_iscont_n(tc, stable_cont) ||
+                MVM_6model_container_iscont_s(tc, stable_cont))) {
+            facts         = &g->facts[ins->operands[0].reg.orig][ins->operands[0].reg.i];
+            /*facts->type   = STABLE(stable_value)->WHAT;*/
             facts->flags |= MVM_SPESH_FACT_RW_CONT;
-            if (IS_CONCRETE(stable_value)) {
-                facts->flags |= MVM_SPESH_FACT_DECONT_CONCRETE;
-                ins->info = MVM_op_get_op(MVM_OP_sp_guardrwconc);
+
+            ins->info = MVM_op_get_op(MVM_OP_sp_guardconc);
+
+            ins->operands = MVM_spesh_alloc(tc, g, 2 * sizeof(MVMSpeshOperand));
+            ins->operands[0] = reg;
+            ins->operands[1].lit_i16 = MVM_spesh_add_spesh_slot(tc, g, (MVMCollectable *)STABLE(stable_cont));
+        } else {
+            if (cs->can_store(tc, stable_cont)) {
+                /* We could do stability testing on rw-ness too, but it's quite
+                 * unlikely we'll have codepaths with a mix of readable and
+                 * writable containers. */
+                facts->flags |= MVM_SPESH_FACT_RW_CONT;
+                if (IS_CONCRETE(stable_value)) {
+                    facts->flags |= MVM_SPESH_FACT_DECONT_CONCRETE;
+                    ins->info = MVM_op_get_op(MVM_OP_sp_guardrwconc);
+                }
+                else {
+                    facts->flags |= MVM_SPESH_FACT_DECONT_TYPEOBJ;
+                    ins->info = MVM_op_get_op(MVM_OP_sp_guardrwtype);
+                }
             }
             else {
-                facts->flags |= MVM_SPESH_FACT_DECONT_TYPEOBJ;
-                ins->info = MVM_op_get_op(MVM_OP_sp_guardrwtype);
+                if (IS_CONCRETE(stable_value)) {
+                    facts->flags |= MVM_SPESH_FACT_DECONT_CONCRETE;
+                    ins->info = MVM_op_get_op(MVM_OP_sp_guardcontconc);
+                }
+                else {
+                    facts->flags |= MVM_SPESH_FACT_DECONT_TYPEOBJ;
+                    ins->info = MVM_op_get_op(MVM_OP_sp_guardconttype);
+                }
             }
+            ins->operands = MVM_spesh_alloc(tc, g, 3 * sizeof(MVMSpeshOperand));
+            ins->operands[0] = reg;
+            ins->operands[1].lit_i16 = MVM_spesh_add_spesh_slot(tc, g, (MVMCollectable *)STABLE(stable_cont));
+            ins->operands[2].lit_i16 = MVM_spesh_add_spesh_slot(tc, g, (MVMCollectable *)STABLE(stable_value));
         }
-        else {
-            if (IS_CONCRETE(stable_value)) {
-                facts->flags |= MVM_SPESH_FACT_DECONT_CONCRETE;
-                ins->info = MVM_op_get_op(MVM_OP_sp_guardcontconc);
-            }
-            else {
-                facts->flags |= MVM_SPESH_FACT_DECONT_TYPEOBJ;
-                ins->info = MVM_op_get_op(MVM_OP_sp_guardconttype);
-            }
-        }
-        ins->operands = MVM_spesh_alloc(tc, g, 3 * sizeof(MVMSpeshOperand));
-        ins->operands[0] = reg;
-        ins->operands[1].lit_i16 = MVM_spesh_add_spesh_slot(tc, g, (MVMCollectable *)STABLE(stable_cont));
-        ins->operands[2].lit_i16 = MVM_spesh_add_spesh_slot(tc, g, (MVMCollectable *)STABLE(stable_value));
     }
     else {
         facts         = &g->facts[ins->operands[0].reg.orig][ins->operands[0].reg.i];
@@ -482,7 +544,7 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
                 g->sf->body.cu->body.hll_config->slurpy_hash_type);
             break;
         case MVM_OP_decont:
-            decont_facts(tc, g,
+            decont_facts(tc, g, ins,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 ins->operands[1].reg.orig, ins->operands[1].reg.i);
             break;
@@ -506,6 +568,72 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 tc->instance->boot_types.BOOTException);
             break;
+        case MVM_OP_getregref_i:
+        case MVM_OP_getlexref_i:
+        case MVM_OP_getregref_i32:
+        case MVM_OP_getlexref_i32:
+        case MVM_OP_getregref_i16:
+        case MVM_OP_getlexref_i16:
+        case MVM_OP_getregref_i8:
+        case MVM_OP_getlexref_i8:
+        case MVM_OP_getregref_u32:
+        case MVM_OP_getlexref_u32:
+        case MVM_OP_getregref_u16:
+        case MVM_OP_getlexref_u16:
+        case MVM_OP_getregref_u8:
+        case MVM_OP_getlexref_u8:
+            create_facts_with_type(tc, g,
+                ins->operands[0].reg.orig, ins->operands[0].reg.i,
+                g->sf->body.cu->body.hll_config->int_lex_ref);
+            break;
+        case MVM_OP_getregref_n:
+        case MVM_OP_getlexref_n:
+        case MVM_OP_getregref_n32:
+        case MVM_OP_getlexref_n32:
+            create_facts_with_type(tc, g,
+                ins->operands[0].reg.orig, ins->operands[0].reg.i,
+                g->sf->body.cu->body.hll_config->num_lex_ref);
+            break;
+        case MVM_OP_getregref_s:
+        case MVM_OP_getlexref_s:
+            create_facts_with_type(tc, g,
+                ins->operands[0].reg.orig, ins->operands[0].reg.i,
+                g->sf->body.cu->body.hll_config->str_lex_ref);
+            break;
+        case MVM_OP_getattrref_i:
+        case MVM_OP_getattrsref_i:
+            create_facts_with_type(tc, g,
+                ins->operands[0].reg.orig, ins->operands[0].reg.i,
+                g->sf->body.cu->body.hll_config->int_attr_ref);
+            break;
+        case MVM_OP_getattrref_n:
+        case MVM_OP_getattrsref_n:
+            create_facts_with_type(tc, g,
+                ins->operands[0].reg.orig, ins->operands[0].reg.i,
+                g->sf->body.cu->body.hll_config->num_attr_ref);
+            break;
+        case MVM_OP_getattrref_s:
+        case MVM_OP_getattrsref_s:
+            create_facts_with_type(tc, g,
+                ins->operands[0].reg.orig, ins->operands[0].reg.i,
+                g->sf->body.cu->body.hll_config->str_attr_ref);
+            break;
+        case MVM_OP_atposref_i:
+            create_facts_with_type(tc, g,
+                ins->operands[0].reg.orig, ins->operands[0].reg.i,
+                g->sf->body.cu->body.hll_config->int_pos_ref);
+            break;
+        case MVM_OP_atposref_n:
+            create_facts_with_type(tc, g,
+                ins->operands[0].reg.orig, ins->operands[0].reg.i,
+                g->sf->body.cu->body.hll_config->num_pos_ref);
+            break;
+        case MVM_OP_atposref_s:
+            create_facts_with_type(tc, g,
+                ins->operands[0].reg.orig, ins->operands[0].reg.i,
+                g->sf->body.cu->body.hll_config->str_pos_ref);
+            break;
+
         case MVM_OP_const_i64:
         case MVM_OP_const_i32:
         case MVM_OP_const_i16:
