@@ -103,7 +103,7 @@ static void optimize_method_lookup(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSp
     if (obj_facts->flags & MVM_SPESH_FACT_KNOWN_TYPE) {
         /* Try to resolve. */
         MVMString *name = MVM_spesh_get_string(tc, g, ins->operands[2]);
-        MVMObject *meth = MVM_6model_find_method_cache_only(tc, obj_facts->type, name);
+        MVMObject *meth = MVM_spesh_try_find_method(tc, obj_facts->type, name);
         if (!MVM_is_null(tc, meth)) {
             /* Could compile-time resolve the method. Add it in a spesh slot. */
             MVMint16 ss = MVM_spesh_add_spesh_slot(tc, g, (MVMCollectable *)meth);
@@ -197,9 +197,8 @@ static void optimize_is_reprid(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshI
 static void optimize_gethow(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *ins) {
     MVMSpeshFacts *obj_facts = MVM_spesh_get_facts(tc, g, ins->operands[1]);
     MVMObject       *how_obj = NULL;
-    if (obj_facts->flags & (MVM_SPESH_FACT_KNOWN_TYPE)) {
-        how_obj = MVM_6model_get_how(tc, STABLE(obj_facts->type));
-    }
+    if (obj_facts->flags & (MVM_SPESH_FACT_KNOWN_TYPE))
+        how_obj = MVM_spesh_try_get_how(tc, obj_facts->type);
     /* There may be other valid ways to get the facts (known value?) */
     if (how_obj) {
         MVMSpeshFacts *how_facts;
@@ -228,6 +227,8 @@ static void optimize_isconcrete(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpesh
         ins->operands[1].lit_i16    = result_facts->value.i;
 
         MVM_spesh_use_facts(tc, g, obj_facts);
+        MVM_spesh_facts_depend(tc, g, result_facts, obj_facts);
+
         obj_facts->usages--;
     }
 }
@@ -245,7 +246,9 @@ static void optimize_exception_ops(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSp
         ins->operands[1].lit_i16 = st->size;
         ins->operands[2].lit_i16 = MVM_spesh_add_spesh_slot(tc, g, (MVMCollectable *)st);
     } else {
+        /*
         MVMSpeshFacts *target_facts;
+        */
 
         /* XXX This currently still causes problems. */
         return;
@@ -597,26 +600,34 @@ static void optimize_decont(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *
         copy_facts(tc, g, ins->operands[0], ins->operands[1]);
     }
     else {
-        MVMSpeshFacts *res_facts;
         if (obj_facts->flags & MVM_SPESH_FACT_KNOWN_TYPE && obj_facts->type) {
             MVMSTable *stable = STABLE(obj_facts->type);
             MVMContainerSpec const *contspec = stable->container_spec;
             if (contspec && contspec->fetch_never_invokes && contspec->spesh) {
                 contspec->spesh(tc, stable, g, bb, ins);
+                MVM_spesh_use_facts(tc, g, obj_facts);
             }
         }
 
-        MVM_spesh_use_facts(tc, g, obj_facts);
-
-        res_facts = MVM_spesh_get_facts(tc, g, ins->operands[0]);
-        if (obj_facts->flags & MVM_SPESH_FACT_KNOWN_DECONT_TYPE) {
-            res_facts->type   = obj_facts->decont_type;
-            res_facts->flags |= MVM_SPESH_FACT_KNOWN_TYPE;
+        if (!MVM_spesh_facts_decont_blocked_by_alias(tc, g, ins)) {
+            MVMSpeshFacts *res_facts = MVM_spesh_get_facts(tc, g, ins->operands[0]);
+            int set_facts = 0;
+            if (obj_facts->flags & MVM_SPESH_FACT_KNOWN_DECONT_TYPE) {
+                res_facts->type   = obj_facts->decont_type;
+                res_facts->flags |= MVM_SPESH_FACT_KNOWN_TYPE;
+                set_facts = 1;
+            }
+            if (obj_facts->flags & MVM_SPESH_FACT_DECONT_CONCRETE) {
+                res_facts->flags |= MVM_SPESH_FACT_CONCRETE;
+                set_facts = 1;
+            }
+            else if (obj_facts->flags & MVM_SPESH_FACT_DECONT_TYPEOBJ) {
+                res_facts->flags |= MVM_SPESH_FACT_TYPEOBJ;
+                set_facts = 1;
+            }
+            if (set_facts)
+                MVM_spesh_facts_depend(tc, g, res_facts, obj_facts);
         }
-        if (obj_facts->flags & MVM_SPESH_FACT_DECONT_CONCRETE)
-            res_facts->flags |= MVM_SPESH_FACT_CONCRETE;
-        else if (obj_facts->flags & MVM_SPESH_FACT_DECONT_TYPEOBJ)
-            res_facts->flags |= MVM_SPESH_FACT_TYPEOBJ;
     }
 }
 
@@ -675,7 +686,7 @@ static void optimize_can_op(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *
     if (MVM_is_null(tc, obj_facts->type))
         can_result = 0; /* VMNull can't have any methods. */
     else
-        can_result = MVM_6model_can_method_cache_only(tc, obj_facts->type, method_name);
+        can_result = MVM_spesh_try_can_method(tc, obj_facts->type, method_name);
 
     if (can_result == -1) {
         return;
@@ -755,7 +766,7 @@ static void optimize_smart_coerce(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpe
 
             return;
         }
-        can_result = MVM_6model_can_method_cache_only(tc, facts->type,
+        can_result = MVM_spesh_try_can_method(tc, facts->type,
                 is_strify ? tc->instance->str_consts.Str : tc->instance->str_consts.Num);
 
         if (can_result == -1) {
@@ -1223,6 +1234,45 @@ static void optimize_extop(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *b
     }
 }
 
+static void optimize_uniprop_ops(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb, MVMSpeshIns *ins) {
+    MVMSpeshFacts *arg1_facts = MVM_spesh_get_facts(tc, g, ins->operands[1]);
+    MVMSpeshFacts *result_facts = MVM_spesh_get_facts(tc, g, ins->operands[0]);
+    if (arg1_facts->flags & MVM_SPESH_FACT_KNOWN_VALUE) {
+        if (ins->info->opcode == MVM_OP_unipropcode) {
+            result_facts->flags |= MVM_SPESH_FACT_KNOWN_VALUE;
+            result_facts->value.i = (MVMint64)MVM_unicode_name_to_property_code(tc, arg1_facts->value.s);
+            ins->info = MVM_op_get_op(MVM_OP_const_i64);
+            ins->operands[1].lit_i64 = result_facts->value.i;
+            arg1_facts->usages--;
+        } else if (ins->info->opcode == MVM_OP_unipvalcode) {
+            MVMSpeshFacts *arg2_facts = MVM_spesh_get_facts(tc, g, ins->operands[2]);
+
+            if (arg2_facts->flags & MVM_SPESH_FACT_KNOWN_VALUE) {
+                result_facts->flags |= MVM_SPESH_FACT_KNOWN_VALUE;
+                result_facts->value.i = (MVMint64)MVM_unicode_name_to_property_value_code(tc, arg1_facts->value.i, arg2_facts->value.s);
+                ins->info = MVM_op_get_op(MVM_OP_const_i64);
+                ins->operands[1].lit_i64 = result_facts->value.i;
+                arg1_facts->usages--;
+                arg2_facts->usages--;
+            }
+        }
+}
+}
+
+/* If something is only kept alive because we log its allocation, kick out
+ * the allocation logging and let the op that creates it die.
+ */
+static void optimize_prof_allocated(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb, MVMSpeshIns *ins) {
+    MVMSpeshFacts *logee_facts = MVM_spesh_get_facts(tc, g, ins->operands[0]);
+    if (logee_facts->usages == 1) {
+        logee_facts->usages = 0;
+        MVM_spesh_manipulate_delete_ins(tc, g, bb, ins);
+        /* This check should always succeed, but just in case ... */
+        if (logee_facts->writer)
+            MVM_spesh_manipulate_delete_ins(tc, g, bb, logee_facts->writer);
+    }
+}
+
 /* Tries to optimize a throwcat instruction. Note that within a given frame
  * (we don't consider inlines here) the throwcat instructions all have the
  * same semantics. */
@@ -1423,6 +1473,7 @@ static void optimize_bb(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb)
         case MVM_OP_prepargs:
             arg_info.cs = g->sf->body.cu->body.callsites[ins->operands[0].callsite_idx];
             arg_info.prepargs_ins = ins;
+            arg_info.prepargs_bb  = bb;
             break;
         case MVM_OP_arg_i:
         case MVM_OP_arg_n:
@@ -1497,6 +1548,10 @@ static void optimize_bb(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb)
         case MVM_OP_objprimspec:
             optimize_objprimspec(tc, g, ins);
             break;
+        case MVM_OP_unipropcode:
+        case MVM_OP_unipvalcode:
+            optimize_uniprop_ops(tc, g, bb, ins);
+            break;
         case MVM_OP_unshift_i:
         case MVM_OP_unshift_n:
         case MVM_OP_unshift_s:
@@ -1725,6 +1780,8 @@ static void second_pass(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb)
 
                MVM_spesh_manipulate_delete_ins(tc, g, bb, ins->prev);
             }
+        } else if (ins->info->opcode == MVM_OP_prof_allocated) {
+            optimize_prof_allocated(tc, g, bb, ins);
         }
 
         ins = ins->next;
