@@ -36,7 +36,7 @@ MVMString * MVM_string_ascii_decode_nt(MVMThreadContext *tc, const MVMObject *re
 
 /* Decodes using a decodestream. Decodes as far as it can with the input
  * buffers, or until a stopper is reached. */
-void MVM_string_ascii_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
+MVMuint32 MVM_string_ascii_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
                                    const MVMint32 *stopper_chars,
                                    MVMDecodeStreamSeparators *seps) {
     MVMint32              count = 0, total = 0;
@@ -45,15 +45,16 @@ void MVM_string_ascii_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
     MVMDecodeStreamBytes *cur_bytes;
     MVMDecodeStreamBytes *last_accept_bytes = ds->bytes_head;
     MVMint32 last_accept_pos, last_was_cr;
+    MVMuint32 reached_stopper;
 
     /* If there's no buffers, we're done. */
     if (!ds->bytes_head)
-        return;
+        return 0;
     last_accept_pos = ds->bytes_head_pos;
 
     /* If we're asked for zero chars, also done. */
     if (stopper_chars && *stopper_chars == 0)
-        return;
+        return 1;
 
     /* Take length of head buffer as initial guess. */
     bufsize = ds->bytes_head->length;
@@ -62,6 +63,7 @@ void MVM_string_ascii_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
     /* Decode each of the buffers. */
     cur_bytes = ds->bytes_head;
     last_was_cr = 0;
+    reached_stopper = 0;
     while (cur_bytes) {
         /* Process this buffer. */
         MVMint32  pos   = cur_bytes == ds->bytes_head ? ds->bytes_head_pos : 0;
@@ -100,10 +102,14 @@ void MVM_string_ascii_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
             last_accept_bytes = cur_bytes;
             last_accept_pos = pos;
             total++;
-            if (stopper_chars && *stopper_chars == total)
+            if (stopper_chars && *stopper_chars == total) {
+                reached_stopper = 1;
                 goto done;
-            if (MVM_string_decode_stream_maybe_sep(tc, seps, codepoint))
+            }
+            if (MVM_string_decode_stream_maybe_sep(tc, seps, codepoint)) {
+                reached_stopper = 1;
                 goto done;
+            }
         }
         cur_bytes = cur_bytes->next;
     }
@@ -118,13 +124,15 @@ void MVM_string_ascii_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
         MVM_free(buffer);
     }
     MVM_string_decodestream_discard_to(tc, ds, last_accept_bytes, last_accept_pos);
+
+    return reached_stopper;
 }
 
 /* Encodes the specified substring to ASCII. Anything outside of ASCII range
  * will become replaced with the supplied replacement, or an exception will be
  * thrown if there isn't one. The result string is NULL terminated, but the
  * specified size is the non-null part. */
-char * MVM_string_ascii_encode_substr(MVMThreadContext *tc, MVMString *str, MVMuint64 *output_size, MVMint64 start, MVMint64 length, MVMString *replacement) {
+char * MVM_string_ascii_encode_substr(MVMThreadContext *tc, MVMString *str, MVMuint64 *output_size, MVMint64 start, MVMint64 length, MVMString *replacement, MVMint32 translate_newlines) {
     /* ASCII is a single byte encoding, but \r\n is a 2-byte grapheme, so we
      * may have to resize as we go. */
     MVMuint32      startu    = (MVMuint32)start;
@@ -142,7 +150,8 @@ char * MVM_string_ascii_encode_substr(MVMThreadContext *tc, MVMString *str, MVMu
         MVM_exception_throw_adhoc(tc, "length out of range");
 
     if (replacement)
-        repl_bytes = MVM_string_ascii_encode_substr(tc, replacement, &repl_length, 0, -1, NULL);
+        repl_bytes = (MVMuint8 *) MVM_string_ascii_encode_substr(tc, replacement,
+            &repl_length, 0, -1, NULL, translate_newlines);
 
     result_alloc = lengthu;
     result = MVM_malloc(result_alloc + 1);
@@ -156,7 +165,7 @@ char * MVM_string_ascii_encode_substr(MVMThreadContext *tc, MVMString *str, MVMu
     else {
         MVMuint32 i = 0;
         MVMCodepointIter ci;
-        MVM_string_ci_init(tc, &ci, str);
+        MVM_string_ci_init(tc, &ci, str, translate_newlines);
         while (MVM_string_ci_has_more(tc, &ci)) {
             MVMCodepoint ord = MVM_string_ci_get_codepoint(tc, &ci);
             if (i == result_alloc) {
@@ -193,11 +202,11 @@ char * MVM_string_ascii_encode_substr(MVMThreadContext *tc, MVMString *str, MVMu
 }
 
 /* Encodes the specified string to ASCII.  */
-char * MVM_string_ascii_encode(MVMThreadContext *tc, MVMString *str, MVMuint64 *output_size) {
-    return MVM_string_ascii_encode_substr(tc, str, output_size, 0, -1, NULL);
+char * MVM_string_ascii_encode(MVMThreadContext *tc, MVMString *str, MVMuint64 *output_size, MVMint32 translate_newlines) {
+    return MVM_string_ascii_encode_substr(tc, str, output_size, 0, -1, NULL, translate_newlines);
 }
 
 /* Encodes the specified string to ASCII not returning length.  */
 char * MVM_string_ascii_encode_any(MVMThreadContext *tc, MVMString *str) {
-    return MVM_string_ascii_encode(tc, str, NULL);
+    return MVM_string_ascii_encode(tc, str, NULL, 0);
 }
