@@ -130,7 +130,7 @@ static MVMNFGTrieNode * twiddle_trie_node(MVMThreadContext *tc, MVMNFGTrieNode *
     /* Free any existing node at next safe point, return the new one. */
     if (current)
         MVM_fixed_size_free_at_safepoint(tc, tc->instance->fsa,
-            sizeof(MVMNGFTrieNodeEntry), current);
+            sizeof(MVMNFGTrieNode), current);
     return new_node;
 }
 static void add_synthetic_to_trie(MVMThreadContext *tc, MVMCodepoint *codes, MVMint32 num_codes, MVMGrapheme32 synthetic) {
@@ -376,3 +376,38 @@ MVMint32 MVM_nfg_is_concat_stable(MVMThreadContext *tc, MVMString *a, MVMString
     /* If either fail quickcheck or have ccc > 0, have to re-normalize. */
     return passes_quickcheck_and_zero_ccc(tc, last_a) && passes_quickcheck_and_zero_ccc(tc, first_b);
 }
+
+/* Free all memory allocated to hold synthetic graphemes. These are global
+ * to a VM instance. */
+void MVM_nfg_destroy(MVMThreadContext *tc) {
+    MVMNFGState *nfg = tc->instance->nfg;
+    MVMint32 i;
+
+    /* Free all synthetics. */
+    if (nfg->synthetics) {
+        size_t used_synths_in_block = nfg->num_synthetics % MVM_SYNTHETIC_GROW_ELEMS;
+        size_t synths_to_free = used_synths_in_block
+            ? nfg->num_synthetics + (MVM_SYNTHETIC_GROW_ELEMS - used_synths_in_block)
+            : nfg->num_synthetics;
+
+        for (i = 0; i < nfg->num_synthetics; i++) {
+            MVM_fixed_size_free(tc, tc->instance->fsa,
+                nfg->synthetics[i].num_combs * sizeof(MVMCodepoint),
+                nfg->synthetics[i].combs);
+            if (nfg->synthetics[i].case_uc != CASE_UNCHANGED)
+                MVM_free(nfg->synthetics[i].case_uc);
+            if (nfg->synthetics[i].case_lc != CASE_UNCHANGED)
+                    MVM_free(nfg->synthetics[i].case_lc);
+            if (nfg->synthetics[i].case_tc != CASE_UNCHANGED)
+                MVM_free(nfg->synthetics[i].case_tc);
+            if (nfg->synthetics[i].case_fc != CASE_UNCHANGED)
+                MVM_free(nfg->synthetics[i].case_fc);
+        }
+
+        MVM_fixed_size_free(tc, tc->instance->fsa,
+            synths_to_free * sizeof(MVMNFGSynthetic),
+            nfg->synthetics);
+    }
+
+    MVM_free(nfg);
+}
