@@ -1,5 +1,5 @@
 /* State kept around for implementing Normal Form Grapheme. The design is such
- * that we can always do lookups without needing to acquire a look. When we
+ * that we can always do lookups without needing to acquire a lock. When we
  * do additions of new synthetics, we must acquire the lock before doing so,
  * and be sure to validate nothing changed. We also must do sufficient copying
  * to ensure that we never break another thread doing a read. Memory to be
@@ -78,3 +78,6 @@ MVMGrapheme32 MVM_nfg_crlf_grapheme(MVMThreadContext *tc);
 MVMNFGSynthetic * MVM_nfg_get_synthetic_info(MVMThreadContext *tc, MVMGrapheme32 synth);
 MVMuint32 MVM_nfg_get_case_change(MVMThreadContext *tc, MVMGrapheme32 codepoint, MVMint32 case_, MVMGrapheme32 **result);
 MVMint32 MVM_nfg_is_concat_stable(MVMThreadContext *tc, MVMString *a, MVMString *b);
+
+/* NFG subsystem cleanup. */
+void MVM_nfg_destroy(MVMThreadContext *tc);
